#  Copyright (c) 2015 - 2022, Intel Corporation
#  SPDX-License-Identifier: BSD-3-Clause
#

"""Implementation for the geopmaccess command line tool.

"""

import sys
import os
import tempfile
import subprocess # nosec
from argparse import ArgumentParser
from dasbus.connection import SystemMessageBus
from dasbus.error import DBusError


class Access:
    """Extension to the GEOPM D-Bus proxy to support the geopmaccess
       command line interface.

    """
    def __init__(self, geopm_proxy):
        """Constructor for Access class

            Args:
                geopm_proxy (dasbus.client.proxy.InterfaceProxy): The
                    dasbus proxy for the GEOPM D-Bus interface.

        """
        try:
            geopm_proxy.PlatformGetGroupAccess
        except DBusError as ee:
            if 'io.github.geopm was not provided' in str(ee):
                err_msg = """The geopm systemd service is not enabled.
    Install geopm service and run 'systemctl start geopm'"""
                raise RuntimeError(err_msg) from ee
            else:
                raise ee
        self._geopm_proxy = geopm_proxy

    def set_group_signals(self, group, signals, is_dry_run, is_force):
        """Call GEOPM D-Bus API to set signal access

        Sets the signal access list for a group while leaving the
        control access list unchanged.  The user must be 'root' to
        perform this operation.  The PlatformGetGroupAccess D-Bus API
        of the io.github.geopm interface is used.

        Args:
            group (str): Unix group name to set access list for.  The
                         call sets the default signal access list if
                         group provided is ''.

            signals (list(str)): List of all signal names that are
                                 allowed for the group or for the
                                 defaults.
            is_dry_run (bool): True to run error checking without file
                               modification
            is_force (bool): True if error checking is disabled


        Raises:
            RuntimeError: The user is not root, the group provided is
                          invalid, or any of the provided signal names
                          are not supported.

        """
        try:
            _, current_controls = self._geopm_proxy.PlatformGetGroupAccess(group)
        except DBusError as ee:
            raise RuntimeError('Failed to read group signal access list for specified group: {}'.format(group)) from ee
        if not is_force:
            signals_supported, _ = self._geopm_proxy.PlatformGetAllAccess()
            signals_requested = set(signals)
            if not signals_requested.issubset(signals_supported):
                missing = ', '.join(sorted(signals_requested.difference(signals_supported)))
                raise RuntimeError(f'Requested access to signals that are not available: {missing}')
        if not is_dry_run:
            try:
                self._geopm_proxy.PlatformSetGroupAccess(group, signals, current_controls)
            except DBusError as ee:
                raise RuntimeError('Failed to set group signal access list, try with sudo or as "root" user (requires CAP_SYS_ADMIN)') from ee

    def set_group_controls(self, group, controls, is_dry_run, is_force):
        """Call GEOPM D-Bus API to set control access

        Sets the control access list for a group while leaving the
        signal access list unchanged.  The user must be 'root' to
        perform this operation.  The PlatformGetGroupAccess D-Bus API
        of the io.github.geopm interface is used.

        Args:
            group (str): Unix group name to set access list for.  The
                         call sets the default control access list if
                         group provided is ''.

            controls (list(str)): List of all control names that are
                                 allowed for the group or for the
                                 defaults.
            is_dry_run (bool): True to run error checking without file
                               modification
            is_force (bool): True if error checking is disabled

        Raises:
            RuntimeError: The user is not root, the group provided is
                          invalid, or any of the provided control names are
                          not supported.

        """
        try:
            current_signals, _ = self._geopm_proxy.PlatformGetGroupAccess(group)
        except DBusError as ee:
            raise RuntimeError('Failed to read group control access list for specified group: {}'.format(group)) from ee
        if not is_force:
            _, controls_supported = self._geopm_proxy.PlatformGetAllAccess()
            controls_requested = set(controls)
            if not controls_requested.issubset(controls_supported):
                missing = ', '.join(sorted(controls_requested.difference(controls_supported)))
                raise RuntimeError(f'Requested access to controls that are not available: {missing}')
        if not is_dry_run:
            try:
                self._geopm_proxy.PlatformSetGroupAccess(group, current_signals, controls)
            except DBusError as ee:
                raise RuntimeError('Failed to set group control access list, try with sudo or as "root" user (requires CAP_SYS_ADMIN)') from ee

    def get_all_signals(self):
        """Call GEOPM D-Bus API and return all supported signal names

        Returns a human readable list of all signals available on the
        platform.  The returned string has one signal name on each
        line, and is in sorted order.  The PlatformGetAllAccess D-Bus API
        of the io.github.geopm interface is used.

        Returns:
            str: All available signals, one on each line, in sorted order

        """
        all_signals, _ = self._geopm_proxy.PlatformGetAllAccess()
        return '\n'.join(sorted(all_signals))

    def get_all_controls(self):
        """Call GEOPM D-Bus API and return all supported control names

        Returns a human readable list of all controls available on the
        system.  The returned string has one control name on each
        line, and is in sorted order.  The PlatformGetAllAccess D-Bus API
        of the io.github.geopm interface is used.

        Returns:
            str: All available controls, one on each line, in sorted order

        """
        _, all_controls = self._geopm_proxy.PlatformGetAllAccess()
        return '\n'.join(sorted(all_controls))

    def get_user_signals(self):
        """Call GEOPM D-Bus API and return user allowed signal names

        Returns a human readable list of user allowed signal names on
        the platform.  The returned string has one signal name on each
        line, and is in sorted order.  The PlatformGetAllAccess D-Bus API
        of the io.github.geopm interface is used.

        Returns:
            str: User allowed signals, one on each line, in sorted order

        """
        user_signals, _ = self._geopm_proxy.PlatformGetUserAccess()
        return '\n'.join(sorted(user_signals))

    def get_user_controls(self):
        """Call GEOPM D-Bus API and return user allowed control names

        Returns a human readable list of user allowed controls on the
        platform.  The returned string has one control name on each
        line, and is in sorted order.  The PlatformGetAllAccess D-Bus API
        of the io.github.geopm interface is used.

        Returns:
            str: User allowed controls, one on each line, in sorted order

        """
        _, user_controls = self._geopm_proxy.PlatformGetUserAccess()
        return '\n'.join(sorted(user_controls))

    def get_group_signals(self, group):
        """Call GEOPM D-Bus API and return the group's signal access list

        Returns a human readable list of the signals that are enabled.
        The returned string has one signal name on each line, and is in sorted order.

        The default signal access list is returned if the group
        provided is the empty string.  If the group provided is not
        empty then the list of signals that are enabled for the
        specified Unix group is returned.

        A user process accessing the GEOPM D-Bus APIs is restricted to
        the combination of the default access list and the access list
        for all groups that the user belongs to.  The results from
        querying a specific Unix group with this method do not reflect
        the default access list that is enabled for all users.

        The Platforms D-Bus API of the io.github.geopm
        interface is called.

        Args:
            group (str): Unix group name to get access list for.  Gets
                         the default signal access list if group
                         provided is '': the empty string.

        Returns:
            str: Access list of signals, one on each line, in sorted order.

        """

        all_signals, _ = self._geopm_proxy.PlatformGetGroupAccess(group)
        return '\n'.join(sorted(all_signals))

    def get_group_controls(self, group):
        """Call GEOPM D-Bus API and return the group's control access list

        Returns a human readable list of the controls that are
        enabled.  The returned string has one control name on each
        line, and is in sorted order.

        The default control access list is returned if the group
        provided is the empty string.  If the group provided is not
        empty then the list of controls that are enabled for the
        specified Unix group is returned.

        A user process accessing the GEOPM D-Bus APIs is restricted to
        the combination of the default access list and the access list
        for all groups that the user belongs to.  The results from
        querying a specific Unix group with this method do not reflect
        the default access list that is enabled for all users.

        The PlatformGetGroupAccess D-Bus API of the io.github.geopm
        interface is called.

        Args:
            group (str): Unix group name to get access list for.  Gets
                         the default control access list if group
                         provided is '': the empty string.

        Returns:
            str: Access list of controls, one on each line, in sorted order.

        """
        _, all_controls = self._geopm_proxy.PlatformGetGroupAccess(group)
        return '\n'.join(sorted(all_controls))

    def _edited_names_helper(self, initial_names):
        default_editor = '/usr/bin/vi'
        if not os.path.exists(default_editor):
            default_editor = '/bin/vi'
        editor = os.environ.get('EDITOR', default_editor)
        with tempfile.NamedTemporaryFile(mode='w+', suffix='geopm-service-access-tmp') as fid:
            path = fid.name
            fid.write(initial_names)
            fid.flush()
            subprocess.run([editor, path], check=True)
            fid.seek(0)
            names = self.read_names(fid)
        return names

    def edited_names(self, is_control, group):
        if is_control:
            initial_names = self.get_group_controls(group)
        else:
            initial_names = self.get_group_signals(group)
        return self._edited_names_helper(initial_names)

    def read_names(self, fid):
        """Returns list names from user

        Parse list of signals or controls from standard input or edit
        existing list.

        """
        return [ll.strip() for ll in fid.readlines() if ll.strip()]

    def run(self, is_write, is_all, is_control, group, is_default, is_delete,
            is_dry_run, is_force, is_edit):
        """Execute geopmaccess command line interface

        The inputs to this method are parsed from the command line
        interface of geopmaccess.  All of the features of the
        geopmaccess tool are implemented with this method.

        Args:

            is_write (bool): True if user requested to write to the
                             access lists, False if it is a read
                             operation.

            is_all (bool): True if the user requested that all
                           available signals or controls be printed as
                           opposed to one of the access lists.

            is_control (bool): True if the user requested to read or
                               write the control access lists.

            group (str): If is_all is not specified, the group
                         determines which access list will be read or
                         written.  If the group is the empty string
                         then the default access list is used,
                         otherwise the parameter specifies the Unix
                         group.
            is_default (bool): True if the default user access list
                               should be printed rather than the calling
                               process' access list.
            is_delete (bool): True to remove an access list file.
            is_dry_run (bool): True to run error checking without file
                               modification
            is_force (bool): True if error checking is disabled

        """
        output = None
        # Determine if user provided -g option
        if group is None:
            is_group = False
            # Empty string is for default access list
            group = ''
        else:
            is_group = True

        if is_all and (is_write or is_edit or is_delete):
            raise RuntimeError('Option -a/--all is not valid when writing a configuration')
        if is_dry_run and (is_edit or is_delete):
            raise RuntimeError('Option -n/--dry-run not valid with -e/--edit or -D/--delete')
        if is_force and is_edit:
            raise RuntimeError('Option -F/--force is not valid with -e/--edit')
        if is_group and (is_default or is_all):
            raise RuntimeError('Option -g/--group is not valid with -u/--default or -a/--all')
        if not (is_edit or is_write or is_delete) and (is_dry_run or is_force):
            raise RuntimeError('-n/--dry-run, and -F/--force not valid when reading')


        names = None
        if is_edit:
            names = self.edited_names(is_control, group)
        elif is_write:
            names = self.read_names(sys.stdin)
        elif is_delete:
            names = []
        if names is not None:
            # Set access list
            if is_control:
                self.set_group_controls(group, names, is_dry_run, is_force)
            else:
                self.set_group_signals(group, names, is_dry_run, is_force)
        else:
            # Get access list
            if is_all:
                if is_control:
                    output = self.get_all_controls()
                else:
                    output = self.get_all_signals()
            elif is_default or is_group:
                if is_control:
                    output = self.get_group_controls(group)
                else:
                    output = self.get_group_signals(group)
            else:
                if is_control:
                    output = self.get_user_controls()
                else:
                    output = self.get_user_signals()
        return output

def main():
    """Access management for the GEOPM Service.

    Command line tool for reading and writing the access management
    lists for the GEOPM Service signals and controls.

    """

    err = 0
    parser = ArgumentParser(description=main.__doc__)
    parser.add_argument('-c', '--controls', dest='controls', action='store_true', default=False,
                        help='Command applies to controls not signals')
    parser_group_uga = parser.add_mutually_exclusive_group(required=False)
    parser_group_uga.add_argument('-u', '--default', dest='default', action='store_true', default=False,
                                  help='Print the default user access list')
    parser_group_uga.add_argument('-g', '--group', dest='group', type=str, default=None,
                                 help='Read or write the access list for a specific Unix GROUP')
    parser_group_uga.add_argument('-a', '--all', dest='all', action='store_true', default=False,
                                  help='Print all signals or controls supported by the service system')
    parser_group_weD = parser.add_mutually_exclusive_group(required=False)
    parser_group_weD.add_argument('-w', '--write', dest='write', action='store_true', default=False,
                                  help='Use standard input to write an access list')
    parser_group_weD.add_argument('-e', '--edit', dest='edit', action='store_true', default=False,
                                  help='Edit an access list using EDITOR environment variable, default vi')
    parser_group_weD.add_argument('-D', '--delete', dest='delete', action='store_true', default=False,
                                  help='Remove an access list for default user or a particular Unix Group')
    parser_group_nF = parser.add_mutually_exclusive_group(required=False)
    parser_group_nF.add_argument('-n', '--dry-run', dest='dry_run', action='store_true', default=False,
                                 help='Do error checking on all user input, but do not modify configuration files')
    parser_group_nF.add_argument('-F', '--force', dest='force', action='store_true', default=False,
                                 help='Write access list without validating GEOPM Service support for names')
    args = parser.parse_args()
    try:
        acc = Access(SystemMessageBus().get_proxy('io.github.geopm',
                                                  '/io/github/geopm'))
        output = acc.run(args.write, args.all, args.controls, args.group,
                         args.default, args.delete, args.dry_run, args.force,
                         args.edit)
        if output:
            print(output)
    except RuntimeError as ee:
        if 'GEOPM_DEBUG' in os.environ:
            # Do not handle exception if GEOPM_DEBUG is set
            raise ee
        sys.stderr.write('Error: {}\n\n'.format(ee))
        err = -1
    return err

if __name__ == '__main__':
    sys.exit(main())



#  Copyright (c) 2015 - 2022, Intel Corporation
#  SPDX-License-Identifier: BSD-3-Clause
#

"""Manage system files used by the geopm service

Provides secure interfaces for manipulating the files in

    /run/geopm-service

that enable the service to be restarted and

   /etc/geopm-service

where the access lists are located.  These interfaces provide
guarantees about the security of these system files, and an
abstraction for updating the contents.

"""

import os
import sys
import stat
import json
import jsonschema
import glob
import psutil
import uuid
import grp
import pwd
import fcntl

from . import pio
from . import schemas

GEOPM_SERVICE_RUN_PATH = '/run/geopm-service'
GEOPM_SERVICE_CONFIG_PATH = '/etc/geopm-service'

GEOPM_SERVICE_RUN_PATH_PERM = 0o711
"""Default permissions for the GEOPM service run path

"""

GEOPM_SERVICE_CONFIG_PATH_PERM = 0o700
"""Default permissions for the GEOPM service config path

"""


def secure_make_dirs(path, perm_mode=0o700):
    """Securely create a directory

    When the path does not exist a directory is created with permissions
    0o700 by default along with all required parent directories.

    The security of this path is verified if the path exists.  An
    existing path is considered to be secure only if all of the
    following conditions are met:

        - The path is a regular directory
        - The path is not a link
        - The path is accessible the the caller
        - The path is a directory owned by the calling process uid
        - The path is a directory owned by the calling process gid
        - The permissions for the directory are 0o700 by default

    If the existing path is determined to be insecure, a warning is
    printed to syslog and the existing file will be renamed to
    `<path>-<UUID>-INVALID` so that it may be audited later, but not
    used.  The warning message will display the reason why the
    directory was not secure.  A new directory with the specified path
    will be created in place of the renamed file.

    This function will simply return when the directory already exists
    and is considered secure.

    Args:
        path (str): The path were the directory is created

    """
    if os.path.exists(path):
        is_valid = True
        renamed_path = f'{path}-{uuid.uuid4()}-INVALID'
        # If it's a link
        if os.path.islink(path):
            is_valid = False
            sys.stderr.write(f'Warning: <geopm-service> {path} is a symbolic link, the link will be renamed to {renamed_path}\n')
            sys.stderr.write(f'Warning: <geopm-service> the symbolic link points to {os.readlink(path)}\n')
        # If it's not a directory
        elif not os.path.isdir(path):
            is_valid = False
            sys.stderr.write(f'Warning: <geopm-service> {path} is not a directory, it will be renamed to {renamed_path}\n')
        # If it's a directory
        else:
            st = os.stat(path)
            # If the permissions are not what we wanted
            set_perm_mode = stat.S_IMODE(st.st_mode)
            if set_perm_mode != perm_mode:
                sys.stderr.write(f'Warning: <geopm-service> {path} has wrong permissions, it will be renamed to {renamed_path}\n')
                sys.stderr.write(f'Warning: <geopm-service> the wrong permissions were {oct(set_perm_mode)}\n')
                is_valid = False
            # If the user owner is not what we wanted
            user_owner = st.st_uid
            if user_owner != os.getuid():
                sys.stderr.write(f'Warning: <geopm-service> {path} has wrong user owner, it will be renamed to {renamed_path}\n')
                sys.stderr.write(f'Warning: <geopm-service> the wrong user owner was {user_owner}\n')
                is_valid = False
            # If the group owner is not what we wanted
            group_owner = st.st_gid
            if group_owner != os.getgid():
                sys.stderr.write(f'Warning: <geopm-service> {path} has wrong group owner, it will be renamed to {renamed_path}\n')
                sys.stderr.write(f'Warning: <geopm-service> the wrong group owner was {group_owner}\n')
                is_valid = False
        # If one of the three above branches revealed an invalid file
        if not is_valid:
            os.rename(path, renamed_path)
            os.mkdir(path, mode=perm_mode)
    # If the path doesn't exist
    else:
        os.mkdir(path, mode=perm_mode)


def secure_make_file(path, contents):
    """Securely and atomically create a file containing a string

    The file is created by first writing the output to a temporary
    file with restricted permissions (mode 0o600) within the same
    directory, and then renaming the file after it is closed.  This
    temporary file with have the path `<path>-<uuid>-tmp`.

    The rename operation is atomic, and will overwrite any existing
    file located in the specified path.  The permissions on the output
    file are mode 0o600 with uid and gid matching the process values.

    Args:
        path (str): The path where the file is created

        contents (str): The contents of the created file

    """
    temp_path = f'{path}-{uuid.uuid4()}-tmp'
    old_mask = os.umask(0o077)
    try:
        with open(os.open(temp_path, os.O_CREAT | os.O_WRONLY, 0o600), 'w') as file:
            file.write(contents)
        os.rename(temp_path, path)
    finally:
        os.umask(old_mask)


def secure_read_file(path):
    """Securely read a file into a string

    The security of this path is verified after the file is opened. An
    existing path is considered to be secure only if all of the
    following conditions are met:

        - The path is a regular file
        - The path is not a link
        - The path is accessible the the caller
        - The path is a file owned by the calling process uid
        - The path is a file owned by the calling process gid
        - The permissions for the file are 0o600

    If the existing path is determined to be insecure, a warning is
    printed to syslog and the existing file will be renamed to
    `<path>-<UUID>-INVALID` so that it may be audited later, but not
    used.  The warning message will display the reason why the file
    was not secure and None is returned.

    If the path points to an existing file that is determined to be
    secure then the contents of the file are returned.

    Args:
        path (str): The path where the file is created

    Returns:
        str: The contents of the file if file was opened, None if couldn't open file

    """
    contents = None
    if is_secure_path(path):
        with open(path) as fid:
            if is_secure_file(path, fid):
                contents = fid.read()
    return contents


def is_secure_path(path):
    """Query if path may be openend safely

    Check if path exists, and refers to a regular file that is not a link.  A
    warning message is printed if the path is a link, a directory or is not a
    regular file and the file is renamed to a path of the form
    `<PATH>-<UUID>-INVALID`.

    Args:
        path (str): The file path

    Returns:
        bool: True if the path is a regular file and not a link

    """
    result = False
    if os.path.exists(path):
        renamed_path = f'{path}-{uuid.uuid4()}-INVALID'
        if os.path.islink(path):
            sys.stderr.write(f'Warning: <geopm-service> {path} is a symbolic link, it will be renamed to {renamed_path}\n')
            sys.stderr.write(f'Warning: <geopm-service> the symbolic link points to {os.readlink(path)}\n')
        elif os.path.isdir(path):
            sys.stderr.write(f'Warning: <geopm-service> {path} is a directory, it will be renamed to {renamed_path}\n')
        elif not stat.S_ISREG(os.stat(path).st_mode):
            sys.stderr.write(f'Warning: <geopm-service> {path} is not a regular file, it will be renamed to {renamed_path}\n')
        else:
            result = True
        if not result:
            os.rename(path, renamed_path)
    else:
        sys.stderr.write(f'Warning: <geopm-service> {path} does not exist\n')
    return result


def is_secure_file(path, fid):
    """Query if file descriptor may be safely read

    After opening a file this function is called to determine if the file
    descriptor is a regular file owned by the calling process user and group
    with restricted permissions (i.e. mode 0o600).  If these conditions are
    not met then a warning message is printed and the file is renamed to a
    path of the form `<PATH>-<UUID>-INVALID`.

    Args:
        path (str): The file path that was passed to open()

        fid (typing.IO): File descriptor returned by open()

    Returns:
        bool: True if regular file with restricted permissions

    """
    result = False
    daemon_uid = os.getuid()
    daemon_gid = os.getgid()
    renamed_path = f'{path}-{uuid.uuid4()}-INVALID'
    warn_msg = f'Warning: <geopm-service> {path} was discovered with invalid permissions, it will be renamed to {renamed_path}\n'
    sess_stat = os.stat(fid.fileno())
    if not stat.S_ISREG(sess_stat.st_mode):
        sys.stderr.write(warn_msg)
        sys.stderr.write(f'Warning: <geopm-service> {path} is not a regular file\n')
    elif stat.S_IMODE(sess_stat.st_mode) != 0o600:
        sys.stderr.write(warn_msg)
        sys.stderr.write(f'Warning: <geopm-service> the wrong permissions were {oct(sess_stat.st_mode)}\n')
    elif sess_stat.st_uid != daemon_uid:
        sys.stderr.write(warn_msg)
        sys.stderr.write(f'Warning: <geopm-service> the wrong user owner was {sess_stat.st_uid}\n')
    elif sess_stat.st_gid != daemon_gid:
        sys.stderr.write(warn_msg)
        sys.stderr.write(f'Warning: <geopm-service> the wrong group owner was {sess_stat.st_gid}\n')
    else:
        result = True
    if not result:
        os.rename(path, renamed_path)
    return result


class ActiveSessions(object):
    """Class that manages the session files for the service

    The state about active sessions opened with geopmd by a client
    processes is stored in files in /run/geopm-service.  These
    files are loaded when the geopmd process starts.  The files are
    modified each time a client opens a session, closes a session,
    requests write permission, or starts a batch server.  The class is
    responsible for managing file access permissions, and atomic file
    creation.

    The session files are stored in JSON format and follow the
    GEOPM_ACTIVE_SESSIONS_SCHEMA documented in the
    ``geopmdpy.schemas`` module.

    """


    def __init__(self, run_path=GEOPM_SERVICE_RUN_PATH):
        """Create an ActiveSessions object that tracks geopmd session files

        The geopmd session files are stored in the directory

            "/run/geopm-service"

        by default, but the user may specify a different path.  The
        creation of an ActiveSessions object will make the directory
        if it does not exist.  This directory is created with
        restricted access permissions (mode: GEOPM_SERVICE_RUN_PATH_PERM).

        If the path points to a symbolic link, the link will be
        renamed and a warning is printed to the syslog.  The directory
        will then be created in its place.  The warning message will
        display the target of the symbolic link and will appear in the
        syslog.

        If the path points to a directory which is not accessible, or
        if the directory is owned by the wrong user or group the
        directory will be renamed and a warning is printed to the
        syslog. The warning message will display the reason why the
        directory was not accessible, or display the permissions or
        user and group of the directory.

        When an ActiveSessions object is created and the directory
        already exists, if the owner of the directory is the geopmd
        user, and the permissions are set to GEOPM_SERVICE_RUN_PATH_PERM
        parsing will proceed.  All files matching the pattern

            "/run/geopm-service/session-*.json"

        will be parsed and verified. These files must conform to the
        session JSON schema, be owned by the geopmd user, and have
        restricted access permissions (mode: 0o600).  Files matching
        the session pattern which do not meet these criterion are not
        parsed, will be renamed, and a warning is printed to the
        syslog.  The warning message will display the user and group
        permissions of the file being renamed.

        Args:
            run_path (str): Optional argument to override the default
                            path to the session files which is
                            "/run/geopm-service"

        Returns:
            ActiveSessions: Object containing all valid sessions data
                            read from the run_path directory

        """
        self._RUN_PATH = run_path
        self._LOCK_PATH = os.path.join(self._RUN_PATH, "CONTROL_LOCK")
        self._M_SHMEM_PREFIX = "geopm-service-batch-buffer-"
        self._M_DEFAULT_FIFO_PREFIX = "batch-status-"
        self._daemon_uid = os.getuid()
        self._daemon_gid = os.getgid()
        self._sessions = dict()
        self._session_schema = json.loads(schemas.GEOPM_ACTIVE_SESSIONS_SCHEMA)
        secure_make_dirs(self._RUN_PATH,
                         perm_mode=GEOPM_SERVICE_RUN_PATH_PERM)

        # Load all session files in the directory
        for sess_path in glob.glob(self._get_session_path('*')):
            self._load_session_file(sess_path)

    def is_client_active(self, client_pid):
        """Query if a Linux PID currently has an active session

        If a session file matching the client_pid exists, but the file
        was not parsed or created by geopmd, then a warning message is
        printed to syslog and the file is renamed.  The warning
        message will display the user and group permissions of the
        file being deleted.  In this case False is returned.

        Args:
            client_pid (int): Linux PID to query

        Returns:
            bool: True if the PID has an open session, False otherwise

        """
        result = client_pid in self._sessions
        session_path = self._get_session_path(client_pid)
        if not result and os.path.isfile(session_path):
            renamed_path = f'{session_path}-{uuid.uuid4()}-INVALID'
            sys.stderr.write(f'Session file exists, but client {client_pid} is not tracked: {session_path} will be moved to {renamed_path}\n')
            os.rename(session_path, renamed_path)
        return result

    def check_client_active(self, client_pid, msg=''):
        """Raise an exception if a PID does not have an active session

        Args:
            client_pid (int): Linux PID to query

            msg (str): The operation that was attempted which requires
                       an active session.  Used in error message

        Raises:
            RuntimeError: Operation not allowed without an open session

        """
        if not self.is_client_active(client_pid):
            raise RuntimeError(f"Operation '{msg}' not allowed without an open session. Client PID: {client_pid}")

    def add_client(self, client_pid, signals, controls, watch_id):
        """Add a new client session to be tracked

        Create a new session file that contains the JSON data provided
        as the call parameters in a format that conforms to the
        session file schema.

        This file will be created atomically such that if this method
        is interrupted before the session file is ready to be used,
        then no file will be present that matches load pattern below.

            ``/run/geopm-service/session-*.json``

        The session file is created without the JSON object property
        "batch_server" specified.  This properties may be modified by the
        set_batch_server() method after the client is added.

        This operation creates the session file atomically, but does
        not modify the session properties nor the session file if the
        client PID already has an open session.

        Args:
            client_pid (int): Linux PID that opened the session

            signals (list(str)): List of signal names that are allowed
                                 to be read by the client

            controls (list(str)): List of control names that are allowed
                                  to be written by the client

            watch_id (int): The watch ID handle returned by GLib for
                            tracking that the PID is active

        """
        if self.is_client_active(client_pid):
            return
        session_data = {'client_pid': client_pid,
                        'reference_count': 1,
                        'signals': list(signals),
                        'controls': list(controls),
                        'watch_id': watch_id}
        jsonschema.validate(session_data, schema=self._session_schema)
        self._sessions[client_pid] = session_data
        self._update_session_file(client_pid)

    def remove_client(self, client_pid):
        """Delete the record of an active session

        Remove the client session file and delete the state associated with
        the client.  Future requests about this client PID will raise an
        exception until another call to add_client() is made, although
        repeated calls to remove_client() for the same PID do not result in an
        error.

        Args:
            client_pid (int): Linux PID that opened the session

        """
        session_path = self._get_session_path(client_pid)
        renamed_path = f'{session_path}-{uuid.uuid4()}-REMOVE'
        sess = None
        try:
            os.rename(session_path, renamed_path)
            content = secure_read_file(renamed_path)
            sess = json.loads(content)
            jsonschema.validate(sess, schema=self._session_schema)
            os.remove(renamed_path)
        except FileNotFoundError:
            pass
        try:
            self._sessions.pop(client_pid)
        except KeyError:
            pass
        return sess

    def get_clients(self):
        """Get list of the client PID values for all active sessions

        Creates a list of the PID values for all active session clients.

        Returns:
            list(int): The Linux PID values for all active clients in sorted order.

        """
        return sorted(list(self._sessions.keys()))

    def get_signals(self, client_pid):
        """Query all signal names that are available

        Creates a list of all signal names that are available to the
        session opened by the client PID.

        Args:
            client_pid (int): Linux PID that opened the session

        Returns:
            list(str): Signal name access list in sorted order for the session

        Raises:
            RuntimeError: Client does not have an open session

        """
        self.check_client_active(client_pid, 'get_signals')
        return sorted(list(self._sessions[client_pid]['signals']))

    def get_controls(self, client_pid):
        """Query all control names that are available

        Creates a list of all control names that are available to the
        session opened by the client PID.

        Args:
            client_pid (int): Linux PID that opened the session

        Returns:
            list(str): Control name access list in sorted order for the session

        Raises:
            RuntimeError: Client does not have an open session

        """
        self.check_client_active(client_pid, 'get_controls')
        return sorted(list(self._sessions[client_pid]['controls']))

    def get_watch_id(self, client_pid):
        """Query for the GLib watch ID for tracking the session lifetime

        The watch ID is not valid in the case where the geopmd process
        restarts.  When a new geopmd process starts it will read files
        that contain watch ID values, however, the set_watch_id()
        method must be called with an updated watch ID for each active
        client session when geopmd starts.

        Args:
            client_pid (int): Linux PID that opened the session

        Returns:
            int: GLib watch ID to track the session lifetime

        Raises:
            RuntimeError: Client does not have an open session

        """
        self.check_client_active(client_pid, 'get_watch_id')
        return self._sessions[client_pid]['watch_id']

    def set_watch_id(self, client_pid, watch_id):
        """Set the GLib watch ID for tracking the session lifetime

        Store the GLib watch ID after registering the callback with
        GLib.timeout_add().

        Args:
            client_pid (int): Linux PID that opened the session

            watch_id (int): GLib watch ID returned by the
                            GLib.timeout_add() method

        Raises:
            RuntimeError: Client does not have an open session

        """
        self.check_client_active(client_pid, 'get_watch_id')
        self._sessions[client_pid]['watch_id'] = watch_id
        self._update_session_file(client_pid)

    def get_reference_count(self, client_pid):
        """Query the reference count for the client session.

        The reference count is basically just an abstraction for the client.
        The client itself cannot have more than a single open session at once,
        since there is a one to one mapping between client pid and open sessions,
        and lifetime of our session is coupled to the lifetime of the connection.

        The reference count is used when in the client process, we have independent components,
        which all may request that same connection as a kind of shared resource.
        We want to provide that same connection as a singleton, and keep track of
        how many independent components have access to that same connection as a resource,
        in order to know when to deallocate that resource when all these components
        have disconnected from the session.

        Args:
            client_pid (int): Linux PID that opened the session

        Returns:
            int: The reference count

        Raises:
            RuntimeError: Client does not have an open session

        """
        self.check_client_active(client_pid, 'get_reference_count')
        return self._sessions[client_pid]['reference_count']

    def set_reference_count(self, client_pid, reference_count):
        """Set the reference count for the session

        This method is used whenever we want to update the reference_count to
        reflect the number of independent components that have connected to the session.

        Args:
            client_pid (int): Linux PID that opened the session

            reference_count (int): The new reference count value

        Raises:
            RuntimeError: Client does not have an open session

        """
        self.check_client_active(client_pid, 'set_reference_count')
        self._sessions[client_pid]['reference_count'] = reference_count
        self._update_session_file(client_pid)

    def increment_reference_count(self, client_pid):
        """Increment the reference_count value by one

        This method is called when a new independent component calls open_session()
        a subsequent time after the connection has already been opened.

        Args:
            client_pid (int): Linux PID that opened the session

        Raises:
            RuntimeError: Client does not have an open session

        """
        self.check_client_active(client_pid, 'increment_reference_count')
        self._sessions[client_pid]['reference_count'] += 1
        self._update_session_file(client_pid)

    def decrement_reference_count(self, client_pid):
        """Decrement the reference_count value by one

        This method is called when an independent component of the client
        calls close_session() and the connection is still open.
        For example when there are multiple independent components using that
        open connection, and one of them wants to close the session, we just
        decrement the reference count, we do not close the connection, as that
        would impact the other independent components which are using the connection.

        Args:
            client_pid (int): Linux PID that opened the session

        Raises:
            RuntimeError: Client does not have an open session

        """
        self.check_client_active(client_pid, 'decrement_reference_count')
        self._sessions[client_pid]['reference_count'] -= 1
        self._update_session_file(client_pid)

    def get_batch_server(self, client_pid):
        """Query the batch server for the client session

        In the case where the client has an open batch server, this
        method returns the Linux PID of the batch server.  If there is
        no active batch server then None is returned.

        Args:
            client_pid (int): Linux PID that opened the session

        Returns:
            int: The Linux PID of the batch server or None

        Raises:
            RuntimeError: Client does not have an open session

        """
        self.check_client_active(client_pid, 'get_batch_server')
        return self._sessions[client_pid].get('batch_server')

    def set_batch_server(self, client_pid, batch_pid):
        """Set the Linux PID of the batch server supporting a client session

        This method is used to store the PID after a batch server is created.

        Args:
            client_pid (int): Linux PID that opened the session

            batch_pid (int): Linux PID of the batch server created for
                             the client session

        Raises:
            RuntimeError: Client does not have an open session

        """
        self.check_client_active(client_pid, 'set_batch_server')
        if 'batch_server' in self._sessions[client_pid]:
            current_server = self._sessions[client_pid]['batch_server']
            raise RuntimeError(f'Client {client_pid} has already started a batch server: {current_server}')
        self._sessions[client_pid]['batch_server'] = batch_pid
        self._update_session_file(client_pid)

    def remove_batch_server(self, client_pid):
        """Remove the record for the batch server supporting a client session

        This method is used to remove the record of a batch server
        after it is closed.

        Args:
            client_pid (int): Linux PID that opened the session

        Raises:
            RuntimeError: Client does not have an open session

        """
        batch_pid = self._sessions[client_pid]['batch_server']
        self._sessions[client_pid].pop('batch_server')

        self.check_client_active(client_pid, 'remove_batch_server')
        self._update_session_file(client_pid)

        signal_shmem_key = self._M_SHMEM_PREFIX + str(batch_pid) + "-signal"
        control_shmem_key = self._M_SHMEM_PREFIX + str(batch_pid) + "-control"
        read_fifo_key = self._M_DEFAULT_FIFO_PREFIX + str(batch_pid) + "-in"
        write_fifo_key = self._M_DEFAULT_FIFO_PREFIX + str(batch_pid) + "-out"
        signal_shmem_path = os.path.join(self._RUN_PATH, signal_shmem_key)
        control_shmem_path = os.path.join(self._RUN_PATH, control_shmem_key)
        read_fifo_path = os.path.join(self._RUN_PATH, read_fifo_key)
        write_fifo_path = os.path.join(self._RUN_PATH, write_fifo_key)

        if (os.path.exists(signal_shmem_path)):
            sys.stderr.write(f'Warning: {signal_shmem_path} file was left over, deleting it now.\n')
            os.unlink(signal_shmem_path)

        if (os.path.exists(control_shmem_path)):
            sys.stderr.write(f'Warning: {control_shmem_path} file was left over, deleting it now.\n')
            os.unlink(control_shmem_path)

        if (os.path.exists(read_fifo_path)):
            sys.stderr.write(f'Warning: {read_fifo_path} file was left over, deleting it now.\n')
            os.unlink(read_fifo_path)

        if (os.path.exists(write_fifo_path)):
            sys.stderr.write(f'Warning: {write_fifo_path} file was left over, deleting it now.\n')
            os.unlink(write_fifo_path)

    def _get_session_path(self, client_pid):
        """Query for the session file path for client PID

        Args:
            client_pid (int): Linux PID that opened the session

        Returns:
            str: Current client's session file path

        """
        return f'{self._RUN_PATH}/session-{client_pid}.json'

    def _update_session_file(self, client_pid):
        """Write the session data to disk for client PID

        Args:
            client_pid (int): Linux PID that opened the session

        """
        sess = self._sessions[client_pid]
        jsonschema.validate(sess, schema=self._session_schema)
        session_path = self._get_session_path(client_pid)
        secure_make_file(session_path, json.dumps(sess))

    def _is_pid_valid(self, pid, file_time):
        """Verify validity of PID

        If the creation time of the process associated with PID
        is newer than or equal to file_time, or if the PID does not
        exist, then the PID is not valid.

        Args:
            pid (int): Linux PID to be verified

            file_time (int): stat.st_ctime to compare against

        Raises:
           psutil.AccessDenied: Insufficient privileges to query for
                                process creation time.

        """
        result = True
        try:
            proc_time = psutil.Process(pid).create_time()
            if proc_time >= file_time:
                # PID has been recycled, return false
                result = False
        except (ValueError, psutil.NoSuchProcess):
            # PID is no longer active
            result = False
        return result

    def _load_session_file(self, sess_path):
        """Load the session file into memory

        If the session file is not valid, it will not be loaded
        and warnings will be printed.  The invalid file will be
        renamed and retained.

        Args:
            sess_path (str): Current client's session file path

        """
        renamed_path = f'{sess_path}-{uuid.uuid4()}-INVALID'
        contents = secure_read_file(sess_path)
        if contents is None:
            return # Invalid JSON return early
        try:
            sess = json.loads(contents)
            jsonschema.validate(sess, schema=self._session_schema)
        except:
            sys.stderr.write(f'Warning: <geopm-service> Invalid JSON file, unable to parse, renamed{sess_path} to {renamed_path} and will ignore\n')
            os.rename(sess_path, renamed_path)
            return # Invalid JSON return early

        sess_stat = os.stat(sess_path)
        file_time = sess_stat.st_ctime
        client_pid = sess['client_pid']
        batch_pid = sess.get('batch_server')
        if self._is_pid_valid(client_pid, file_time):
            # Valid session; verify batch server
            if batch_pid is not None and \
               self._is_pid_valid(batch_pid, file_time) == False:
                sess.pop('batch_server')
            self._sessions[client_pid] = dict(sess)
            self._update_session_file(client_pid)
        else:
           os.rename(sess_path, renamed_path)


class AccessLists(object):
    """Class that manages the access list files

    """
    def __init__(self, config_path=GEOPM_SERVICE_CONFIG_PATH):
        self._CONFIG_PATH = config_path
        secure_make_dirs(self._CONFIG_PATH)
        self._DEFAULT_ACCESS = '0.DEFAULT_ACCESS'
        self._pio = pio

    def _validate_group(self, group):
        if group is None or group == '':
            group = self._DEFAULT_ACCESS
        else:
            group = str(group)
            if group[0].isdigit():
                raise RuntimeError('Linux group name cannot begin with a digit: group = "{}"'.format(group))
            try:
                grp.getgrnam(group)
            except KeyError:
                raise RuntimeError('Linux group is not defined: group = "{}"'.format(group))
        return group

    def _read_allowed(self, path):
        contents = secure_read_file(path)
        result = []
        if contents is not None:
            result = [line.strip() for line in contents.splitlines()
                      if line.strip() and not line.strip().startswith('#')]
        return result

    def _write_allowed(self, path, allowed):
        allowed.append('')
        contents = '\n'.join(allowed)
        secure_make_file(path, contents)

    def _filter_valid_signals(self, signals):
        signals = set(signals)
        all_signals = self._pio.signal_names()
        return list(signals.intersection(all_signals))

    def _filter_valid_controls(self, controls):
        controls = set(controls)
        all_controls = self._pio.control_names()
        return list(controls.intersection(all_controls))

    def _get_user_groups(self, user):
        user_gid = pwd.getpwnam(user).pw_gid
        all_gid = os.getgrouplist(user, user_gid)
        return [grp.getgrgid(gid).gr_name for gid in all_gid]

    def get_group_access(self, group):
        """Get the signal and control access lists

        Read the list of allowed signals and controls for the
        specified group.  If the group is None or the empty string
        then the default lists of allowed signals and controls are
        returned.

        The values are securely read from files located in
        /etc/geopm-service using the secure_read_file() interface.

        If no secure file exist for the specified group, then two
        empty lists are returned.

        Args:
            group (str): Name of group

        Returns:
            list(str)), list(str): Signal and control allowed lists, in sorted order.

        Raises:
            RuntimeError: The group name is not valid on the system.

        """
        group = self._validate_group(group)
        group_dir = os.path.join(self._CONFIG_PATH, group)
        if os.path.isdir(group_dir):
            secure_make_dirs(group_dir)
            path = os.path.join(group_dir, 'allowed_signals')
            signals = self._read_allowed(path)
            signals = self._filter_valid_signals(signals)
            signals = sorted(signals)
            path = os.path.join(group_dir, 'allowed_controls')
            controls = self._read_allowed(path)
            controls = self._filter_valid_controls(controls)
            controls = sorted(controls)
        else:
            signals = []
            controls = []
        return signals, controls

    def set_group_access(self, group, allowed_signals, allowed_controls):
        """Set signals and controls in the allowed lists

        Write the list of allowed signals and controls for the
        specified group.  If the group is None or the empty string
        then the default lists of allowed signals and controls are
        updated.

        The values are securely written atomically to files located in
        /etc/geopm-service using the secure_make_dirs() and
        secure_make_file() interfaces.

        Args:
            group (str): Name of group

            allowed_signals (list(str)): Signal names that are allowed

            allowed_controls (list(str)): Control names that are allowed

        Raises:
            RuntimeError: The group name is not valid on the system.

        """
        group = self._validate_group(group)
        group_dir = os.path.join(self._CONFIG_PATH, group)
        secure_make_dirs(group_dir,
                         perm_mode=GEOPM_SERVICE_CONFIG_PATH_PERM)
        path = os.path.join(group_dir, 'allowed_signals')
        self._write_allowed(path, allowed_signals)
        path = os.path.join(group_dir, 'allowed_controls')
        self._write_allowed(path, allowed_controls)

    def get_user_access(self, user):
        """Get the list of all of the signals and controls that are
        accessible to the specified user.

        Returns the default access lists that apply to all non-root
        users if the empty string is provided.

        All available signals and controls are returned if the caller
        specifies the user name 'root'.  A RuntimeError is
        raised if the user does not exist on the system.

        When a user requests a signal or control through one of the
        other PlatformService methods, they are restricted to the
        union of the default allowed lists and the allowed lists for
        all Unix groups that the user belongs to.  These combined
        lists are what this method returns.

        Args:
            user (str): Which Unix user name to query; if the empty
                        string is provided, the default allowed list
                        is returned.

        Returns:
            list(str), list(str): Signal and control allowed lists, in sorted order.

        Raises:
            RuntimeError: The user does not exist.

        """
        if user == 'root':
            return self.get_all_access()
        user_groups = []
        if user != '':
            try:
                user_groups = self._get_user_groups(user)
            except KeyError as e:
                raise RuntimeError("Specified user '{}' does not exist.".format(user))
        user_groups.append('') # Default access list
        signal_set = set()
        control_set = set()
        for group in user_groups:
            signals, controls = self.get_group_access(group)
            signal_set.update(signals)
            control_set.update(controls)
        signals = sorted(signal_set)
        controls = sorted(control_set)

        return signals, controls

    def get_all_access(self):
        """Get all of the signals and controls that the service supports.

        Returns the list of all signals and controls supported by the
        service.  The lists returned are independent of the access
        controls; therefore, calling get_all_access() is equivalent
        to calling get_user_access('root').

        Returns:
            list(str), list(str): All supported signals and controls, in sorted order.

        """
        return self._pio.signal_names(), self._pio.control_names()


class WriteLock(object):
    """Class for interacting with control lock file

    This class provides the interface to query and set the PID that owns the
    GEOPM Service control write lock.  The state of this lock is stored in the
    file path:

        /run/geopm-service/CONTROL_LOCK

    This file is empty when the lock is free, and contains the PID of the
    controlling process when the lock is held.  The class manages the advisory
    lock to serialize any attempts read or write the file.  Additionally the
    class checks that the lock file is a regular file with restricted
    permissions.  Maniuplations of the control lock file should be done
    exclusively with the WriteLock object to insure that the advisory lock is
    effective.

    """
    def __init__(self, run_path=GEOPM_SERVICE_RUN_PATH):
        """Set up initial state

        The WriteLock must be used within a context manager.  Use a
        non-default run_path only for testing purposes.  This constructor will
        securely create the run_path if it does not exist.

        Args:
            run_path: Directory to create control lock within

        """
        self._RUN_PATH = run_path
        self._LOCK_PATH = os.path.join(self._RUN_PATH, "CONTROL_LOCK")
        self._fid = None
        secure_make_dirs(self._RUN_PATH,
                         perm_mode=GEOPM_SERVICE_RUN_PATH_PERM)

    def __enter__(self):
        """Enter context management for interacting with write lock

        Securely open the lock file for editing and hold exclusive advisory
        lock.  The file pointer is rewound after the lock is held.  Two
        attempts will be made to securely open the file and any insecure files
        will be renamed.

        """
        self._rpc_lock()
        old_mask = os.umask(0o077)
        try:
            trial_count = 0
            while self._fid is None and trial_count < 2:
                if os.path.exists(self._LOCK_PATH):
                    # Rename existing file if it is not secure
                    is_secure_path(self._LOCK_PATH)
                self._fid = open(self._LOCK_PATH, 'a+')
                if not is_secure_file(self._LOCK_PATH, self._fid):
                    # File is insecure and was renamed, try again
                    self._fid.close()
                    self._fid = None
                trial_count += 1
        finally:
            os.umask(old_mask)
        if self._fid == None:
            self._rpc_unlock()
            raise RuntimeError('Unable to open write lock securely after two tries')
        # Advisory lock protects against simultanious multi-process
        # modifications to the file, although we expect only one geopmd
        # process using this class.
        fcntl.lockf(self._fid, fcntl.LOCK_EX)
        self._fid.seek(0)
        return self

    def __exit__(self, type, value, traceback):
        """Exit context management for interacting with write lock

        Release the advisory lock and close the file.

        """
        if self._fid is not None:
            fcntl.lockf(self._fid, fcntl.LOCK_UN)
            self._fid.close()
            self._rpc_unlock()


    _is_file_active = False
    @classmethod
    def _rpc_lock(cls):
        """Lock to protect against nested use of the write lock within one process

        """
        if cls._is_file_active:
            raise RuntimeError('Attempt to modify control lock file while file lock is held by the same process')
        cls._is_file_active = True

    @classmethod
    def _rpc_unlock(cls):
        """Unlock single process mutex

        """
        if not cls._is_file_active:
            raise RuntimeError('Unlock failed, file is not active')
        cls._is_file_active = False

    def try_lock(self, pid=None):
        """Get the PID that holds the lock or set lock

        Returns the PID that currently holds the lock.  If the user specifies
        a PID and the lock is not held by another process, then the lock will
        be assigned, and the input pid will be returned.  The user must check
        the return value to determine if a request to assign the lock was
        successful, an error is not raised if the lock is held by another PID.

        None is returned if the write lock is not held by any active session
        and the user does not specify a pid.

        Args:
            pid (int): The PID to assign the lock

        Returns:
            int: The PID of the session that holds the write lock upon return
                 or None if the write lock is not held

        """
        contents = self._read()
        if contents == '':
            if pid is not None:
                self._write(pid)
            file_pid = pid
        else:
            file_pid = int(contents)
        return file_pid

    def unlock(self, pid):
        """Release the write lock

        Release the write lock from ownership by a specified PID.  If the lock
        is not currently assigned to the specified PID, then a RuntimeError is
        raised.

        Args:
            pid (int): The PID that the lock is assigned to

        Raises:
            RuntimeError: The specified PID does not currently hold the lock

        """
        contents = self._read()
        if contents == '':
            raise RuntimeError('Lock is not held by any PID')
        else:
            file_pid = int(contents)
            if file_pid == pid:
                self._fid.truncate(0)
            else:
                raise RuntimeError(f'Lock is held by another PID: {file_pid}')

    def _read(self):
        """Read the contents of the lock file

        Rewinds the file pointer after the read.

        Returns:
           str: contents of the file

        Raises:
           RuntimeError: WriteLock object not used within a context manager

        """

        if self._fid is None:
            raise RuntimeError('The WriteLock object must be used within a context manager')
        contents = self._fid.readline(64)
        self._fid.seek(0)
        return contents

    def _write(self, pid):
        """Write a PID to a lock file

        Rewinds the file pointer after the write.

        Args:
           pid (int):  PID to write into the lock file

        """
        pid = int(pid)
        self._fid.truncate(0)
        self._fid.write(str(pid))
        self._fid.flush()
        self._fid.seek(0)



#  Copyright (c) 2015 - 2022, Intel Corporation
#  SPDX-License-Identifier: BSD-3-Clause
#

"""Implementation for the geopmsession command line tool

"""

import sys
import os
import time
import math
from argparse import ArgumentParser
from dasbus.connection import SystemMessageBus
from dasbus.error import DBusError
from . import topo
from . import pio
from . import runtime


class Session:
    """Object responsible for creating a GEOPM service session

    This object's run() method is the main entry point for
    geopmsession command line tool.  The inputs to run() are derived
    from the command line options provided by the user.

    The Session object depends on the RequestQueue object to parse the
    input request buffer from the user.  The Session object also
    depends on the runtime.TimedLoop object when executing a periodic
    read session.

    """

    def __init__(self, geopm_proxy):
        """Constructor for Session class

        Args:
            geopm_proxy (dasbus.client.proxy.InterfaceProxy): The
                dasbus proxy for the GEOPM D-Bus interface.

        Raises:
            RuntimeError: The geopm systemd service is not running

        """
        try:
            geopm_proxy.PlatformOpenSession
        except DBusError as ee:
            if 'io.github.geopm was not provided' in str(ee):
                err_msg = """The geopm systemd service is not enabled.
    Install geopm service and run 'systemctl start geopm'"""
                raise RuntimeError(err_msg) from ee
            else:
                raise ee
        self._geopm_proxy = geopm_proxy

    def format_signals(self, signals, signal_format):
        """Format a list of signal values for printing

        Args:
            signals (list(float)): Values to be printed

            signal_format (list(int)): The geopm::format_string_e enum
                                       value describing the formatting
                                       for each signal provided.

        Returns:
            str: Ready-to-print line of formatted values

        """
        if len(signals) != len(signal_format):
            raise RuntimeError(
                'Number of signal values does not match the number of requests')
        result = [pio.format_signal(ss, ff) for (ss, ff) in
                  zip(signals, signal_format)]
        return '{}\n'.format(','.join(result))

    def run_read(self, requests, duration, period, out_stream):
        """Run a read mode session

        Use the GEOPM D-Bus interface to periodically read the
        requested signals. A line of text will be printed to the
        output stream for each period of time.  The line will contain
        a comma separated list of the read values, one for each
        request.

        Only one read of the requests is made if the period is zero.
        If the period is non-zero then the requested signals are read
        periodically.  The first read is immediate, and then a delay
        of period seconds elapses until the next sample is made.
        These periodic reads are executed and printed until the
        duration of time specified has been met or exceeded.

        Args:
            requests (ReadRequestQueue): Request object parsed from
                                         user input.

            duration (float): The user specified minimum length of time
                              for the samples to span in units of
                              seconds.

            period (float): The user specified period between samples
                            in units of seconds.

            out_stream (typing.IO): Object with write() method where output
                               will be printed (typically sys.stdout).

        """
        num_period = 0
        if period != 0:
            num_period = math.ceil(duration / period)
        signal_handles = []
        for name, dom, dom_idx in requests:
            if not name.startswith('SERVICE::'):
                name = f'SERVICE::{name}'
            signal_handles.append(pio.push_signal(name, dom, dom_idx))

        for sample_idx in runtime.TimedLoop(period, num_period):
            pio.read_batch()
            signals = [pio.sample(handle) for handle in signal_handles]
            line = self.format_signals(signals, requests.get_formats())
            out_stream.write(line)

    def check_read_args(self, run_time, period):
        """Check that the run time and period are valid for a read session

        Args:
            run_time (float): Time duration of the session open in
                              seconds.

            period (float): The user specified period between samples
                            in units of seconds.

        Raises:
            RuntimeError: The period is greater than the total time or
                          either is negative.

        """
        if period > run_time:
            raise RuntimeError('Specified a period that is greater than the total run time')
        if period > 86400:
            raise RuntimeError('Specified a period greater than 24 hours')
        if period < 0.0 or run_time < 0.0:
            raise RuntimeError('Specified a negative run time or period')

    def run(self, run_time, period,
            request_stream=sys.stdin, out_stream=sys.stdout):
        """"Create a GEOPM session with values parsed from the command line

        The implementation for the geopmsession command line tool.
        The inputs to this method are derived from the parsed command
        line provided by the user.

        Args:
            run_time (float): Time duration of the session in seconds.

            period (float): Time interval for each line of output for.
                            Value must be zero for a write mode
                            session.

            request_stream (typing.IO): Input from user describing the
                                   requests to read.

            out_stream (typing.IO): Stream where output from will be
                               printed.

        """
        requests = ReadRequestQueue(request_stream, self._geopm_proxy)
        self.check_read_args(run_time, period)
        self.run_read(requests, run_time, period, out_stream)


class RequestQueue:
    """Object derived from user input that provides request information

    The geopmsession command line tool parses requests for reading
    from standard input.  The RequestQueue object holds the logic for
    parsing the input stream upon construction.  The resulting object
    may be iterated upon to retrieve the requested signals that the
    user would like to read.  The Request object also provides the
    enum used to format signal values into strings.

    """
    def __init__(self):
        raise NotImplementedError('RequestQueue class is an abstract base class for ReadRequestQueue')

    def __iter__(self):
        """Iterate over list of requests

        """
        yield from self._requests

    def get_names(self):
        """Get the signal or control names from each request

        Strip off "SERVICE::" prefix from any names in request queue,
        so that the name can be passed to the proxy.

        Returns:
            list(str): The name of the signal or control associated
                           with each user request.

        """
        key = 'SERVICE::'
        return [rr[0] if not rr[0].startswith(key) else rr[0].replace(key, '', 1)
                for rr in self._requests]

    def iterate_stream(self, request_stream):
        """Iterate over a stream of requests

        This is a generator function that will filter out comment
        lines and trailing white space from the input stream.  It can
        be used to iterate over a request stream from the user for
        either a read or write mode session.

        Args:
            request_stream (typing.IO): Stream containing requests from
                                   user.

        Returns:
            typing.Generator[str]: Iterate over filtered lines of the
                                   request_stream

        """
        for line in request_stream:
            line = line.strip()
            if line == '':
                break
            if not line.startswith('#'):
                yield line


class ReadRequestQueue(RequestQueue):
    def __init__(self, request_stream, geopm_proxy):
        """Constructor for ReadRequestQueue object

        Args:
            request_stream (typing.IO): Input from user describing the
                                   requests to read signals.

            geopm_proxy (dasbus.client.proxy.InterfaceProxy): The
                dasbus proxy for the GEOPM D-Bus interface.

        Raises:
            RuntimeError: The geopm systemd service is not running

        """
        self._requests = self.parse_requests(request_stream)
        self._formats = self.query_formats(self.get_names(), geopm_proxy)

    def parse_requests(self, request_stream):
        """Parse input stream and return list of read requests

        Parse a user supplied stream into a list of tuples
        representing read requests.  The tuples are of the form
        (signal_name, domain_type, domain_idx) and are parsed one from
        each line of the stream.  Each of the values in the stream is
        separated by white space.

        Each signal_name should match one of the signal names provided
        by the service.  The domain_type is specified as the name
        string, i.e one of the following strings: "board", "package",
        "core", "cpu", "memory", "package_integrated_memory", "nic",
        "package_integrated_nic", "gpu", "package_integrated_gpu".
        The domain index is a positive integer indexing the specific
        domain.

        Args:
            request_stream (typing.IO): Input stream to parse for read
                                   requests

        Returns:
            list((str, int, int)): List of request tuples. Each
                                   request comprises a signal name,
                                   domain type, and domain index.

        Raises:
            RuntimeError: Line from stream does not split into three
                          words and is also not a comment or empty
                          line.

        """
        requests = []
        for line in self.iterate_stream(request_stream):
            words = line.split()
            if len(words) != 3:
                raise RuntimeError('Read request must be three words: "{}"'.format(line))
            try:
                signal_name = words[0]
                domain_type = topo.domain_type(words[1])
                domain_idx = int(words[2])
            except (RuntimeError, ValueError):
                raise RuntimeError('Unable to convert values into a read request: "{}"'.format(line))
            requests.append((signal_name, domain_type, domain_idx))
        if len(requests) == 0:
            raise RuntimeError('Empty request stream.')
        return requests

    def query_formats(self, signal_names, geopm_proxy):
        """Call the GEOPM D-Bus API to get the format type for each signal name

        Returns a list of geopm::string_format_e integers that determine
        how each signal value is formatted as a string.

        Args:
            signal_names (list(str)): List of signal names to query.

            geopm_proxy (dasbus.client.proxy.InterfaceProxy): The
                dasbus proxy for the GEOPM D-Bus interface.

        Returns:
            list(int): List of geopm::string_format_e integers, one
            for each signal name in input list.

        """

        try:
            result = [info[4] for info in
                      geopm_proxy.PlatformGetSignalInfo(signal_names)]
        except DBusError as ee:
            if 'io.github.geopm was not provided' in str(ee):
                err_msg = """The geopm systemd service is not enabled.
    Install geopm service and run 'systemctl start geopm'"""
                raise RuntimeError(err_msg) from ee
            else:
                raise ee
        return result

    def get_formats(self):
        """Get formatting enum values for the parsed read requests

        Returns:
            list(int): The geopm::string_format_e enum value for each
                       read request.

        """
        return self._formats

def main():
    """Command line interface for the geopm service batch read features.

    This command can be used to read signals by opening a session with
    the geopm service.

    """
    err = 0
    parser = ArgumentParser(description=main.__doc__)
    parser.add_argument('-t', '--time', dest='time', type=float, default=0.0,
                        help='Total run time of the session to be opened in seconds')
    parser.add_argument('-p', '--period', dest='period', type=float, default = 0.0,
                        help='When used with a read mode session reads all values out periodically with the specified period in seconds')
    args = parser.parse_args()
    try:
        sess = Session(SystemMessageBus().get_proxy('io.github.geopm',
                                                    '/io/github/geopm'))
        sess.run(args.time, args.period)
    except RuntimeError as ee:
        if 'GEOPM_DEBUG' in os.environ:
            # Do not handle exception if GEOPM_DEBUG is set
            raise ee
        sys.stderr.write('Error: {}\n\n'.format(ee))
        err = -1
    return err

if __name__ == '__main__':
    sys.exit(main())



#
#  Copyright (c) 2015 - 2022, Intel Corporation
#  SPDX-License-Identifier: BSD-3-Clause
#

"""The geopm daemon python package

"""

import os

try:
    from geopmdpy.version import __version__
except ImportError:
    try:
        # Look for VERSION file in git repository
        file_path = os.path.abspath(__file__)
        src_version_path = os.path.join(
                           os.path.dirname(
                           os.path.dirname(
                           os.path.dirname(file_path))),
                           'VERSION')
        with open(src_version_path) as fid:
            __version__ = fid.read().strip()
    except IOError:
        __version__ = '0.0.0'



#
#  Copyright (c) 2015 - 2022, Intel Corporation
#  SPDX-License-Identifier: BSD-3-Clause
#


"""The GEOPM DBus API enables a client to make measurements from the
hardware platform and set hardware control parameters.  Fine grained
permissions management for both measurements (signals) and controls is
configurable by system administrators."""
_module_doc=__doc__

try:
    from docstring_parser import google
    _use_docstring_parser = True
except ImportError as ex:
    _use_docstring_parser = False
    _use_docstring_parser_err = str(ex)

def geopm_dbus_xml(TopoService=None, PlatformService=None):
    do_parse_docs = _use_docstring_parser
    if TopoService is None or PlatformService is None:
        do_parse_docs = False

    format_string ="""\
<node xmlns:doc="http://www.freedesktop.org/dbus/1.0/doc.dtd">
  <interface name="io.github.geopm">
    <doc:doc>
      <doc:description>
        <doc:para>{module_doc}
        </doc:para>
      </doc:description>
    </doc:doc>
    <method name="TopoGetCache">
      <arg direction="out" name="result" type="s">
        <doc:doc>
          <doc:summary>{TopoGetCache_returns_description}
          </doc:summary>
        </doc:doc>
      </arg>
      <doc:doc>
        <doc:description>
          <doc:summary>{TopoGetCache_short_description}
          </doc:summary>
          <doc:para>{TopoGetCache_long_description}
          </doc:para>
        </doc:description>
      </doc:doc>
    </method>
    <method name="PlatformGetGroupAccess">
      <arg direction="in" name="group" type="s">
        <doc:doc>
          <doc:summary>{PlatformGetGroupAccess_params0_description}
          </doc:summary>
        </doc:doc>
      </arg>
      <arg direction="out" name="access_lists" type="(asas)">
        <doc:doc>
          <doc:summary>{PlatformGetGroupAccess_returns_description}
          </doc:summary>
        </doc:doc>
      </arg>
      <doc:doc>
        <doc:description>
          <doc:summary>{PlatformGetGroupAccess_short_description}
          </doc:summary>
          <doc:para>{PlatformGetGroupAccess_long_description}
          </doc:para>
        </doc:description>
      </doc:doc>
    </method>
    <method name="PlatformSetGroupAccess">
      <arg direction="in" name="group" type="s">
        <doc:doc>
          <doc:summary>{PlatformSetGroupAccess_params0_description}
          </doc:summary>
        </doc:doc>
      </arg>
      <arg direction="in" name="allowed_signals" type="as">
        <doc:doc>
          <doc:summary>{PlatformSetGroupAccess_params1_description}
          </doc:summary>
        </doc:doc>
      </arg>
      <arg direction="in" name="allowed_controls" type="as">
        <doc:doc>
          <doc:summary>{PlatformSetGroupAccess_params2_description}
          </doc:summary>
        </doc:doc>
      </arg>
      <doc:doc>
        <doc:description>
          <doc:summary>{PlatformSetGroupAccess_short_description}
          </doc:summary>
          <doc:para>{PlatformSetGroupAccess_long_description}
          </doc:para>
        </doc:description>
      </doc:doc>
    </method>
    <method name="PlatformGetUserAccess">
      <arg direction="out" name="access_lists" type="(asas)">
        <doc:doc>
          <doc:summary>{PlatformGetUserAccess_returns_description}
          </doc:summary>
        </doc:doc>
      </arg>
      <doc:doc>
        <doc:description>
          <doc:summary>{PlatformGetUserAccess_short_description}
          </doc:summary>
          <doc:para>{PlatformGetUserAccess_long_description}
          </doc:para>
        </doc:description>
      </doc:doc>
    </method>
    <method name="PlatformGetAllAccess">
      <arg direction="out" name="access_lists" type="(asas)">
        <doc:doc>
          <doc:summary>{PlatformGetAllAccess_returns_description}
          </doc:summary>
        </doc:doc>
      </arg>
      <doc:doc>
        <doc:description>
          <doc:summary>{PlatformGetAllAccess_short_description}
          </doc:summary>
          <doc:para>{PlatformGetAllAccess_long_description}
          </doc:para>
        </doc:description>
      </doc:doc>
    </method>
    <method name="PlatformGetSignalInfo">
      <arg direction="in" name="signal_names" type="as">
        <doc:doc>
          <doc:summary>{PlatformGetSignalInfo_params0_description}
          </doc:summary>
        </doc:doc>
      </arg>
      <arg direction="out" name="info" type="a(ssiiii)">
        <doc:doc>
          <doc:summary>{PlatformGetSignalInfo_returns_description}
          </doc:summary>
        </doc:doc>
      </arg>
      <doc:doc>
        <doc:description>
          <doc:summary>{PlatformGetSignalInfo_short_description}
          </doc:summary>
          <doc:para>{PlatformGetSignalInfo_long_description}
          </doc:para>
        </doc:description>
      </doc:doc>
    </method>
    <method name="PlatformGetControlInfo">
      <arg direction="in" name="control_names" type="as">
        <doc:doc>
          <doc:summary>{PlatformGetControlInfo_params0_description}
          </doc:summary>
        </doc:doc>
      </arg>
      <arg direction="out" name="info" type="a(ssi)">
        <doc:doc>
          <doc:summary>{PlatformGetControlInfo_returns_description}
          </doc:summary>
        </doc:doc>
      </arg>
      <doc:doc>
        <doc:description>
          <doc:summary>{PlatformGetControlInfo_short_description}
          </doc:summary>
          <doc:para>{PlatformGetControlInfo_long_description}
          </doc:para>
        </doc:description>
      </doc:doc>
    </method>
    <method name="PlatformLockControl">
     <doc:doc>
        <doc:description>
          <doc:summary>{PlatformLockControl_short_description}
          </doc:summary>
          <doc:para>{PlatformLockControl_long_description}
          </doc:para>
        </doc:description>
      </doc:doc>
    </method>
    <method name="PlatformUnlockControl">
     <doc:doc>
        <doc:description>
          <doc:summary>{PlatformUnlockControl_short_description}
          </doc:summary>
          <doc:para>{PlatformUnlockControl_long_description}
          </doc:para>
        </doc:description>
      </doc:doc>
    </method>
    <method name="PlatformOpenSession">
     <doc:doc>
        <doc:description>
          <doc:summary>{PlatformOpenSession_short_description}
          </doc:summary>
          <doc:para>{PlatformOpenSession_long_description}
          </doc:para>
        </doc:description>
      </doc:doc>
    </method>
    <method name="PlatformCloseSession">
     <doc:doc>
        <doc:description>
          <doc:summary>{PlatformCloseSession_short_description}
          </doc:summary>
          <doc:para>{PlatformCloseSession_long_description}
          </doc:para>
        </doc:description>
      </doc:doc>
    </method>
    <method name="PlatformCloseSessionAdmin">
      <arg direction="in" name="client_pid" type="i">
        <doc:doc>
          <doc:summary>{PlatformCloseSessionAdmin_params0_description}
          </doc:summary>
        </doc:doc>
      </arg>
      <doc:doc>
        <doc:description>
          <doc:summary>{PlatformCloseSessionAdmin_short_description}
          </doc:summary>
          <doc:para>{PlatformCloseSessionAdmin_long_description}
          </doc:para>
        </doc:description>
      </doc:doc>
    </method>
    <method name="PlatformStartBatch">
      <arg direction="in" name="signal_config" type="a(iis)">
        <doc:doc>
          <doc:summary>{PlatformStartBatch_params0_description}
          </doc:summary>
        </doc:doc>
      </arg>
      <arg direction="in" name="control_config" type="a(iis)">
        <doc:doc>
          <doc:summary>{PlatformStartBatch_params1_description}
          </doc:summary>
        </doc:doc>
      </arg>
      <arg direction="out" name="batch" type="(is)">
        <doc:doc>
          <doc:summary>{PlatformStartBatch_returns_description}
          </doc:summary>
        </doc:doc>
      </arg>
      <doc:doc>
        <doc:description>
          <doc:summary>{PlatformStartBatch_short_description}
          </doc:summary>
          <doc:para>{PlatformStartBatch_long_description}
          </doc:para>
        </doc:description>
      </doc:doc>
    </method>
    <method name="PlatformStopBatch">
      <arg direction="in" name="server_pid" type="i">
        <doc:doc>
          <doc:summary>{PlatformStopBatch_params0_description}
          </doc:summary>
        </doc:doc>
      </arg>
      <doc:doc>
        <doc:description>
          <doc:summary>{PlatformStopBatch_short_description}
          </doc:summary>
          <doc:para>{PlatformStopBatch_long_description}
          </doc:para>
        </doc:description>
      </doc:doc>
    </method>
    <method name="PlatformReadSignal">
      <arg direction="in" name="signal_name" type="s">
        <doc:doc>
          <doc:summary>{PlatformReadSignal_params0_description}
          </doc:summary>
        </doc:doc>
      </arg>
      <arg direction="in" name="domain" type="i">
        <doc:doc>
          <doc:summary>{PlatformReadSignal_params1_description}
          </doc:summary>
        </doc:doc>
      </arg>
      <arg direction="in" name="domain_idx" type="i">
        <doc:doc>
          <doc:summary>{PlatformReadSignal_params2_description}
          </doc:summary>
        </doc:doc>
      </arg>
      <arg direction="out" name="sample" type="d">
        <doc:doc>
          <doc:summary>{PlatformReadSignal_returns_description}
          </doc:summary>
        </doc:doc>
      </arg>
      <doc:doc>
        <doc:description>
          <doc:summary>{PlatformReadSignal_short_description}
          </doc:summary>
          <doc:para>{PlatformReadSignal_long_description}
          </doc:para>
        </doc:description>
      </doc:doc>
    </method>
    <method name="PlatformWriteControl">
      <arg direction="in" name="control_name" type="s">
        <doc:doc>
          <doc:summary>{PlatformWriteControl_params0_description}
          </doc:summary>
        </doc:doc>
      </arg>
      <arg direction="in" name="domain" type="i">
        <doc:doc>
          <doc:summary>{PlatformWriteControl_params1_description}
          </doc:summary>
        </doc:doc>
      </arg>
      <arg direction="in" name="domain_idx" type="i">
        <doc:doc>
          <doc:summary>{PlatformWriteControl_params2_description}
          </doc:summary>
        </doc:doc>
      </arg>
      <arg direction="in" name="setting" type="d">
        <doc:doc>
          <doc:summary>{PlatformWriteControl_params3_description}
          </doc:summary>
        </doc:doc>
      </arg>
      <doc:doc>
        <doc:description>
          <doc:summary>{PlatformWriteControl_short_description}
          </doc:summary>
          <doc:para>{PlatformWriteControl_long_description}
          </doc:para>
        </doc:description>
      </doc:doc>
    </method>
  </interface>
</node>"""

    if do_parse_docs:
        TopoGetCache = google.parse(TopoService.get_cache.__doc__)
        PlatformGetGroupAccess = google.parse(PlatformService.get_group_access.__doc__)
        PlatformSetGroupAccess = google.parse(PlatformService.set_group_access.__doc__)
        PlatformGetUserAccess = google.parse(PlatformService.get_user_access.__doc__)
        PlatformGetAllAccess = google.parse(PlatformService.get_all_access.__doc__)
        PlatformGetSignalInfo = google.parse(PlatformService.get_signal_info.__doc__)
        PlatformGetControlInfo = google.parse(PlatformService.get_control_info.__doc__)
        PlatformLockControl = google.parse(PlatformService.lock_control.__doc__)
        PlatformUnlockControl = google.parse(PlatformService.unlock_control.__doc__)
        PlatformOpenSession = google.parse(PlatformService.open_session.__doc__)
        PlatformCloseSession = google.parse(PlatformService.close_session.__doc__)
        PlatformCloseSessionAdmin = google.parse(PlatformService.close_session_admin.__doc__)
        PlatformStartBatch = google.parse(PlatformService.start_batch.__doc__)
        PlatformStopBatch = google.parse(PlatformService.stop_batch.__doc__)
        PlatformReadSignal = google.parse(PlatformService.read_signal.__doc__)
        PlatformWriteControl = google.parse(PlatformService.write_control.__doc__)
        result = format_string.format(
            module_doc=_module_doc,
            TopoGetCache_returns_description=TopoGetCache.returns.description,
            TopoGetCache_short_description=TopoGetCache.short_description,
            TopoGetCache_long_description=TopoGetCache.long_description,
            PlatformGetGroupAccess_params0_description=PlatformGetGroupAccess.params[0].description,
            PlatformGetGroupAccess_returns_description=PlatformGetGroupAccess.returns.description,
            PlatformGetGroupAccess_short_description=PlatformGetGroupAccess.short_description,
            PlatformGetGroupAccess_long_description=PlatformGetGroupAccess.long_description,
            PlatformSetGroupAccess_params0_description=PlatformSetGroupAccess.params[0].description,
            PlatformSetGroupAccess_params1_description=PlatformSetGroupAccess.params[1].description,
            PlatformSetGroupAccess_params2_description=PlatformSetGroupAccess.params[2].description,
            PlatformSetGroupAccess_short_description=PlatformGetGroupAccess.short_description,
            PlatformSetGroupAccess_long_description=PlatformGetGroupAccess.long_description,
            PlatformGetUserAccess_returns_description=PlatformGetUserAccess.returns.description,
            PlatformGetUserAccess_short_description=PlatformGetUserAccess.short_description,
            PlatformGetUserAccess_long_description=PlatformGetUserAccess.long_description,
            PlatformGetAllAccess_returns_description=PlatformGetAllAccess.returns.description,
            PlatformGetAllAccess_short_description=PlatformGetAllAccess.short_description,
            PlatformGetAllAccess_long_description=PlatformGetAllAccess.long_description,
            PlatformGetSignalInfo_params0_description=PlatformGetSignalInfo.params[0].description,
            PlatformGetSignalInfo_returns_description=PlatformGetSignalInfo.returns.description,
            PlatformGetSignalInfo_short_description=PlatformGetSignalInfo.short_description,
            PlatformGetSignalInfo_long_description=PlatformGetSignalInfo.long_description,
            PlatformGetControlInfo_params0_description=PlatformGetControlInfo.params[0].description,
            PlatformGetControlInfo_returns_description=PlatformGetControlInfo.returns.description,
            PlatformGetControlInfo_short_description=PlatformGetControlInfo.short_description,
            PlatformGetControlInfo_long_description=PlatformGetControlInfo.long_description,
            PlatformLockControl_short_description=PlatformLockControl.short_description,
            PlatformLockControl_long_description=PlatformLockControl.long_description,
            PlatformUnlockControl_short_description=PlatformUnlockControl.short_description,
            PlatformUnlockControl_long_description=PlatformUnlockControl.long_description,
            PlatformOpenSession_short_description=PlatformOpenSession.short_description,
            PlatformOpenSession_long_description=PlatformOpenSession.long_description,
            PlatformCloseSession_params0_description=PlatformCloseSession.params[0].description,
            PlatformCloseSession_short_description=PlatformCloseSession.short_description,
            PlatformCloseSession_long_description=PlatformCloseSession.long_description,
            PlatformCloseSessionAdmin_params0_description=PlatformCloseSessionAdmin.params[0].description,
            PlatformCloseSessionAdmin_short_description=PlatformCloseSessionAdmin.short_description,
            PlatformCloseSessionAdmin_long_description=PlatformCloseSessionAdmin.long_description,
            PlatformStartBatch_params0_description=PlatformStartBatch.params[1].description,
            PlatformStartBatch_params1_description=PlatformStartBatch.params[2].description,
            PlatformStartBatch_returns_description=PlatformStartBatch.returns.description,
            PlatformStartBatch_short_description=PlatformStartBatch.short_description,
            PlatformStartBatch_long_description=PlatformStartBatch.long_description,
            PlatformStopBatch_params0_description=PlatformStopBatch.params[0].description,
            PlatformStopBatch_short_description=PlatformStopBatch.short_description,
            PlatformStopBatch_long_description=PlatformStopBatch.long_description,
            PlatformReadSignal_params0_description=PlatformReadSignal.params[0].description,
            PlatformReadSignal_params1_description=PlatformReadSignal.params[1].description,
            PlatformReadSignal_params2_description=PlatformReadSignal.params[2].description,
            PlatformReadSignal_returns_description=PlatformReadSignal.returns.description,
            PlatformReadSignal_short_description=PlatformReadSignal.short_description,
            PlatformReadSignal_long_description=PlatformReadSignal.long_description,
            PlatformWriteControl_params0_description=PlatformWriteControl.params[0].description,
            PlatformWriteControl_params1_description=PlatformWriteControl.params[1].description,
            PlatformWriteControl_params2_description=PlatformWriteControl.params[2].description,
            PlatformWriteControl_params3_description=PlatformWriteControl.params[3].description,
            PlatformWriteControl_short_description=PlatformWriteControl.short_description,
            PlatformWriteControl_long_description=PlatformWriteControl.long_description)
    else:
        result = _remove_doc(format_string)
    return result


def _remove_doc(xml_str):
    """Remove DBus XML documentation from string input

    """
    result = []
    do_add = True
    for line in xml_str.splitlines():
        if '<doc:doc>' in line:
            do_add = False
        if do_add:
            result.append(line)
        if '</doc:doc>' in line:
            do_add = True
    return '\n'.join(result)


if __name__ == '__main__':
    from . import service

    if not _use_docstring_parser:
        raise ImportError(f'{_use_docstring_parser_err}: The docstring_parser python package is required to use CLI')

    print("""<!DOCTYPE node PUBLIC
    "-//freedesktop//DTD D-BUS Object Introspection 1.0//EN"
     "http://www.freedesktop.org/standards/dbus/1.0/introspect.dtd">
""")
    print(geopm_dbus_xml(service.TopoService, service.PlatformService))



#
#  Copyright (c) 2015 - 2022, Intel Corporation
#  SPDX-License-Identifier: BSD-3-Clause
#

"""Json schemas used by geopmdpy

GEOPM_ACTIVE_SESSIONS_SCHEMA:

.. code-block:: json

    {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "id": "https://geopm.github.io/active_sessions.schema.json",
        "title": "ActiveSession",
        "type": "object",
        "properties": {
          "client_pid": {
            "type": "integer"
          },
          "reference_count": {
            "type": "integer",
            "minimum": 0
          },
          "signals": {
            "type": "array",
            "items": {
              "type": "string"
            }
          },
          "controls": {
            "type": "array",
            "items": {
              "type": "string"
            }
          },
          "watch_id": {
            "type": "integer"
          },
          "batch_server": {
            "type": "integer"
          }
        },
        "required": ["client_pid", "signals", "controls"],
        "additionalProperties": false
    }


"""

GEOPM_ACTIVE_SESSIONS_SCHEMA = """
    {
        "$schema": "http://json-schema.org/draft-04/schema#",
        "id": "https://geopm.github.io/active_sessions.schema.json",
        "title": "ActiveSession",
        "type": "object",
        "properties": {
          "client_pid": {
            "type": "integer"
          },
          "reference_count": {
            "type": "integer",
            "minimum": 0
          },
          "signals": {
            "type": "array",
            "items": {
              "type": "string"
            }
          },
          "controls": {
            "type": "array",
            "items": {
              "type": "string"
            }
          },
          "watch_id": {
            "type": "integer"
          },
          "batch_server": {
            "type": "integer"
          }
        },
        "required": ["client_pid", "signals", "controls"],
        "additionalProperties": false
    }
"""



#
#  Copyright (c) 2015 - 2022, Intel Corporation
#  SPDX-License-Identifier: BSD-3-Clause
#

'''The gffi module provides a wrapper around the cffi interface

This module enables a single cffi.FFI() object to be used throughout
all of the GEOPM python modules and also enables us to enforce that
the libgeopmpolicy.so dynamic library is opened prior to libgeopmd.so.
This is required because libgeopmd.so allocates static objects that
depend on static objects defined in libgeopmpolicy.so (in particular
the geopm::ApplicationSampler).

'''

import cffi

'''gffi is the global FFI object used by all geopm python modules

'''
gffi = cffi.FFI()

def get_dl_geopmd():
    '''Get the FFILibrary instance for libgeopmd.so

    Returns:
        FFILibrary: Object used to call functions defined in
                    libgeopmd.so

    '''
    global _dl_geopmd
    if type(_dl_geopmd) is OSError:
        raise _dl_geopmd
    return _dl_geopmd

def get_dl_geopmpolicy():
    '''Get the FFILibrary instance for libgeopmpolicy.so

    Returns:
        FFILibrary: Object used to call functions defined in
                    libgeopmpolicy.so

    '''
    global _dl_geopmpolicy
    if type(_dl_geopmpolicy) is OSError:
        raise _dl_geopmpolicy
    return _dl_geopmpolicy

# Enforce load order of libgeopmpolicy.so and libgeopmd.so
try:
    _dl_geopmpolicy = gffi.dlopen('libgeopmpolicy.so.1',
                                  gffi.RTLD_GLOBAL|gffi.RTLD_LAZY)
except OSError as err:
    _dl_geopmpolicy = err

try:
    _dl_geopmd =  gffi.dlopen('libgeopmd.so.1',
                              gffi.RTLD_GLOBAL|gffi.RTLD_LAZY)
except OSError as err:
    _dl_geopmd = err



#
#  Copyright (c) 2015 - 2022, Intel Corporation
#  SPDX-License-Identifier: BSD-3-Clause
#


import sys
from . import gffi

gffi.gffi.cdef("""
enum geopm_error_e {
    GEOPM_ERROR_RUNTIME = -1,
    GEOPM_ERROR_LOGIC = -2,
    GEOPM_ERROR_INVALID = -3,
    GEOPM_ERROR_FILE_PARSE = -4,
    GEOPM_ERROR_LEVEL_RANGE = -5,
    GEOPM_ERROR_NOT_IMPLEMENTED = -6,
    GEOPM_ERROR_PLATFORM_UNSUPPORTED = -7,
    GEOPM_ERROR_MSR_OPEN = -8,
    GEOPM_ERROR_MSR_READ = -9,
    GEOPM_ERROR_MSR_WRITE = -10,
    GEOPM_ERROR_AGENT_UNSUPPORTED = -11,
    GEOPM_ERROR_AFFINITY = -12,
    GEOPM_ERROR_NO_AGENT = -13,
};

void geopm_error_message(int err, char *msg, size_t size);

""")
_dl = gffi.get_dl_geopmd()

ERROR_RUNTIME = _dl.GEOPM_ERROR_RUNTIME
ERROR_LOGIC = _dl.GEOPM_ERROR_LOGIC
ERROR_INVALID = _dl.GEOPM_ERROR_INVALID
ERROR_FILE_PARSE = _dl.GEOPM_ERROR_FILE_PARSE
ERROR_LEVEL_RANGE = _dl.GEOPM_ERROR_LEVEL_RANGE
ERROR_NOT_IMPLEMENTED = _dl.GEOPM_ERROR_NOT_IMPLEMENTED
ERROR_PLATFORM_UNSUPPORTED = _dl.GEOPM_ERROR_PLATFORM_UNSUPPORTED
ERROR_MSR_OPEN = _dl.GEOPM_ERROR_MSR_OPEN
ERROR_MSR_READ = _dl.GEOPM_ERROR_MSR_READ
ERROR_MSR_WRITE = _dl.GEOPM_ERROR_MSR_WRITE
ERROR_AGENT_UNSUPPORTED = _dl.GEOPM_ERROR_AGENT_UNSUPPORTED
ERROR_AFFINITY = _dl.GEOPM_ERROR_AFFINITY
ERROR_NO_AGENT = _dl.GEOPM_ERROR_NO_AGENT

def message(err_number):
    """Return the error message associated with the error code.  Positive
    error codes are interpreted as system error numbers, and
    negative error codes are interpreted as GEOPM error numbers.

    Args:
        err_number (int): Error code to be interpreted.

    Returns:
        str: Error message associated with error code.

    """
    global _dl

    name_max = 1024
    result_cstr = gffi.gffi.new("char[]", name_max)
    _dl.geopm_error_message(err_number, result_cstr, name_max)
    return gffi.gffi.string(result_cstr).decode()



#
#  Copyright (c) 2015 - 2022, Intel Corporation
#  SPDX-License-Identifier: BSD-3-Clause
#


from . import gffi
from . import error


gffi.gffi.cdef("""
enum geopm_domain_e {
    GEOPM_DOMAIN_INVALID = -1,
    GEOPM_DOMAIN_BOARD = 0,
    GEOPM_DOMAIN_PACKAGE = 1,
    GEOPM_DOMAIN_CORE = 2,
    GEOPM_DOMAIN_CPU = 3,
    GEOPM_DOMAIN_MEMORY = 4,
    GEOPM_DOMAIN_PACKAGE_INTEGRATED_MEMORY = 5,
    GEOPM_DOMAIN_NIC = 6,
    GEOPM_DOMAIN_PACKAGE_INTEGRATED_NIC = 7,
    GEOPM_DOMAIN_GPU = 8,
    GEOPM_DOMAIN_PACKAGE_INTEGRATED_GPU = 9,
    GEOPM_DOMAIN_GPU_CHIP = 10,
    GEOPM_NUM_DOMAIN = 11,
};

int geopm_topo_num_domain(int domain_type);

int geopm_topo_domain_idx(int domain_type,
                          int cpu_idx);

int geopm_topo_num_domain_nested(int inner_domain,
                                 int outer_domain);

int geopm_topo_domain_nested(int inner_domain,
                             int outer_domain,
                             int outer_idx,
                             size_t num_domain_nested,
                             int *domain_nested);

int geopm_topo_domain_name(int domain_type,
                           size_t domain_name_max,
                           char *domain_name);

int geopm_topo_domain_type(const char *domain_name);

int geopm_topo_create_cache(void);
""")
_dl = gffi.get_dl_geopmd()

DOMAIN_INVALID = _dl.GEOPM_DOMAIN_INVALID
DOMAIN_BOARD = _dl.GEOPM_DOMAIN_BOARD
DOMAIN_PACKAGE = _dl.GEOPM_DOMAIN_PACKAGE
DOMAIN_CORE = _dl.GEOPM_DOMAIN_CORE
DOMAIN_CPU = _dl.GEOPM_DOMAIN_CPU
DOMAIN_MEMORY = _dl.GEOPM_DOMAIN_MEMORY
DOMAIN_PACKAGE_INTEGRATED_MEMORY = _dl.GEOPM_DOMAIN_PACKAGE_INTEGRATED_MEMORY
DOMAIN_NIC = _dl.GEOPM_DOMAIN_NIC
DOMAIN_PACKAGE_INTEGRATED_NIC = _dl.GEOPM_DOMAIN_PACKAGE_INTEGRATED_NIC
DOMAIN_GPU = _dl.GEOPM_DOMAIN_GPU
DOMAIN_PACKAGE_INTEGRATED_GPU = _dl.GEOPM_DOMAIN_PACKAGE_INTEGRATED_GPU
DOMAIN_GPU_CHIP = _dl.GEOPM_DOMAIN_GPU_CHIP
NUM_DOMAIN = _dl.GEOPM_NUM_DOMAIN

def num_domain(domain):
    """Get the number of domains available on the system of a specific
    domain type.  If the domain is valid, but there are no domains of
    that type on the system, the return value is zero.  Invalid domain
    specificiation will result in a raised exception.

    Args:
        domain (int or str): The domain type from one of topo.DOMAIN_*
            integer values or a domain name string.

    Returns:
        int: The number of domains of the specified type available on
            the system.

    """
    global _dl
    domain = domain_type(domain)
    result = _dl.geopm_topo_num_domain(domain)
    if result < 0:
        raise RuntimeError("geopm_topo_num_domain() failed: {}".format(
            error.message(result)))
    return result

def domain_idx(domain, cpu_idx):
    """Get the index of the domain that is local to a specific Linux
    logical CPU.  The return value will be greater than or equal to
    zero and less than the value returned by num_domain(domain).  An
    exception is raised if either input is out of range or invalid.

    Args:
        domain (int or str): The domain type from one of topo.DOMAIN_*
            integer values or a domain name string.
        cpu_idx (int): The Linux logical CPU that is associated with
            the returned domain type.

    Returns:
        int: Domain index associated with the specified logical CPU.

    """
    global _dl
    domain = domain_type(domain)
    result = _dl.geopm_topo_domain_idx(domain, cpu_idx)
    if result < 0:
        raise RuntimeError("geopm_topo_domain_idx() failed: {}".format(
            error.message(result)))
    return result

def domain_nested(inner_domain, outer_domain, outer_idx):
    """Get a list of all inner domains nested within a specified outer
    domain.  An exception is raised if the inner domain is not within
    the outer domain or if the index is out of range.

    Args:
        inner_domain (int or str): The inner domain type from one of
            topo.DOMAIN_* integer values or a domain name string.
        outer_domain (int or str): The outer domain type from one of
            topo.DOMAIN_* integer values or a domain name string.
        outer_idx (int): Index of the outer domain that is queried.

    Returns:
        list[int]: The inner domain indices that are contained
            within the specified outer domain.

    """
    global _dl
    inner_domain = domain_type(inner_domain)
    outer_domain = domain_type(outer_domain)
    num_domain_nested = _dl.geopm_topo_num_domain_nested(inner_domain, outer_domain)
    if (num_domain_nested < 0):
        raise RuntimeError("geopm_topo_num_domain_nested() failed: {}".format(
            error.message(num_domain_nested)))
    domain_nested_carr = gffi.gffi.new("int[]", num_domain_nested)
    _dl.geopm_topo_domain_nested(inner_domain, outer_domain, outer_idx,
                                 num_domain_nested, domain_nested_carr)
    result = []
    for dom in domain_nested_carr:
        result.append(dom)
    return result

def domain_name(domain):
    """Get the domain name corresponding to the domain type specified.  If
    domain is a string, a check is done that the string is a valid
    domain type and then the input string is returned.  This is the
    inverse of the domain_type() function.

    Args:
        domain (int or str): The domain type from one of topo.DOMAIN_*
            integer values or a domain name string.

    Returns:
        str: Domain name string associated with input.

    """
    global _dl
    domain = domain_type(domain)
    name_max = 1024
    result_cstr = gffi.gffi.new("char[]", name_max)
    err = _dl.geopm_topo_domain_name(domain, name_max, result_cstr)
    if err < 0:
        raise RuntimeError("geopm_topo_domain_name() failed: {}".format(
                           error.message(err)))
    return gffi.gffi.string(result_cstr).decode()

def domain_type(domain):
    """Returns the domain type that is associated with the provided domain
    string.  If domain is of integer type, a check is done that it is
    a valid topo.DOMAIN_* value and then the input value is returned.
    This is the inverse of the domain_name() function.  If domain is a
    string that does not match any of the valid domain names, or if domain
    is an integer that is out of range an exception is raised.

    Args:
        domain (str or int): A domain name string or the domain type
            from one of topo.DOMAIN_* integer values.

    Returns:
        int: A domain type matching one of the topo.DOMAIN_* values
            associated with the input.

    """
    if type(domain) is int:
        if domain >= 0 and domain < NUM_DOMAIN:
            result = domain
        else:
            raise RuntimeError("domain_type is out of range: {}".format(domain))
    else:
        global _dl
        domain_cstr = gffi.gffi.new("char[]", domain.encode())
        result = _dl.geopm_topo_domain_type(domain_cstr)
        if result < 0:
            raise RuntimeError("geopm_topo_domain_type() failed: {}".format(
                error.message(result)))
    return result

def create_cache():
    """Create a cache file for the platform topology if one does not
    exist.  This cache file will be used by any calls to the other
    functions in the topo module as well as any use of the GEOPM
    runtime.  File permissions of the cache file are set to
    "-rw-r--r--", i.e. 644. The path for the cache file is
    /run/geopm-service/geopm-topo-cache.  If this file is older than
    the last boot it will be regenerated.  If the file exists but has
    improper permissions is will be regenerated.  If the file has been
    created since the last boot with the correct permissions, no
    operation will be performed.  To force the creation of a new cache
    file call os.unlink('/run/geopm-service/geopm-topo-cache') prior
    to calling this function.

    """
    global _dl
    err = _dl.geopm_topo_create_cache()
    if err < 0:
        raise RuntimeError("geopm_topo_create_cache() failed: {}".format(
            error.message(err)))



#
#  Copyright (c) 2015 - 2022, Intel Corporation
#  SPDX-License-Identifier: BSD-3-Clause
#

"""Module containing the implementations of the DBus interfaces
exposed by geopmd."""

import os
import sys
import signal
import pwd
import grp
import shutil
import psutil
import uuid
import gi
from gi.repository import GLib

from . import pio
from . import topo
from . import dbus_xml
from . import system_files
from dasbus.connection import SystemMessageBus
try:
    from dasbus.server.interface import accepts_additional_arguments
except ImportError as ee:
    err_msg = '''dasbus version greater than 1.5 required:
    https://github.com/rhinstaller/dasbus/pull/57'''
    raise ImportError(err_msg) from ee


class PlatformService(object):
    """Provides the concrete implementation for all of the GEOPM DBus
    interfaces that use PlatformIO.  This class is used by the
    GEOPMService class that maps DBus interfaces to their
    implementations.

    """
    def __init__(self):
        """PlatformService constructor that initializes all private members.

        """
        self._pio = pio
        self._RUN_PATH = system_files.GEOPM_SERVICE_RUN_PATH
        self._SAVE_DIR = 'SAVE_FILES'
        self._WATCH_INTERVAL_SEC = 1
        self._active_sessions = system_files.ActiveSessions()
        self._access_lists = system_files.AccessLists()
        for client_pid in self._active_sessions.get_clients():
            is_active = self.check_client(client_pid)
            if is_active:
                watch_id = self._watch_client(client_pid)
                self._active_sessions.set_watch_id(client_pid, watch_id)
        with system_files.WriteLock(self._RUN_PATH) as lock:
            write_pid = lock.try_lock()
            if write_pid is not None and not self._active_sessions.is_client_active(write_pid):
                self._close_session_write(lock, write_pid)

    def get_group_access(self, group):
        """Get the signal and control access lists

        Read the list of allowed signals and controls for the
        specified group.  If the group is None or the empty string
        then the default lists of allowed signals and controls are
        returned.

        The values are securely read from files located in
        /etc/geopm-service using the secure_read_file() interface.

        If no secure file exist for the specified group, then two
        empty lists are returned.

        Args:
            group (str): Name of group

        Returns:
            list(str), list(str): Signal and control allowed lists, both in sorted order.

        Raises:
            RuntimeError: The group name is not valid on the system.

        """
        return self._access_lists.get_group_access(group)

    def set_group_access(self, group, allowed_signals, allowed_controls):
        """Set signals and controls in the allowed lists

        Write the list of allowed signals and controls for the
        specified group.  If the group is None or the empty string
        then the default lists of allowed signals and controls are
        updated.

        The values are securely written atomically to files located in
        /etc/geopm-service using the secure_make_dirs() and
        secure_make_file() interfaces.

        Args:
            group (str): Name of group

            allowed_signals (list(str)): Signal names that are allowed

            allowed_controls (list(str)): Control names that are allowed

        Raises:
            RuntimeError: The group name is not valid on the system.

        """
        self._access_lists.set_group_access(group, allowed_signals, allowed_controls)

    def get_user_access(self, user):
        """Get the list of all of the signals and controls that are
        accessible to the specified user.

        Returns the default access lists that apply to all non-root
        users if the empty string is provided.

        All available signals and controls are returned if the caller
        specifies the user name 'root'.  A RuntimeError is
        raised if the user does not exist on the system.

        When a user requests a signal or control through one of the
        other PlatformService methods, they are restricted to the
        union of the default allowed lists and the allowed lists for
        all Unix groups that the user belongs to.  These combined
        lists are what this method returns.

        Args:
            user (str): Which Unix user name to query; if the empty
                        string is provided, the default allowed list
                        is returned.

        Returns:
            list(str), list(str): Signal and control allowed lists, both in sorted order.

        Raises:
            RuntimeError: The user does not exist.

        """
        return self._access_lists.get_user_access(user)

    def get_all_access(self):
        """Get all of the signals and controls that the service supports.

        Returns the list of all signals and controls supported by the
        service.  The lists returned are independent of the access
        controls; therefore, calling get_all_access() is equivalent
        to calling get_user_access('root').

        Returns:
            list(str), list(str): All supported signals and controls, in sorted order.

        """
        return self._access_lists.get_all_access()

    def get_signal_info(self, signal_names):
        """For each specified signal name, return a tuple of information.

        The caller provides a list of signal names.  This method returns
        a list of the same length which consists of a tuple of information
        for each signal.

        The method will raise a RuntimeError if any of the requested
        signal names are not supported by the service.

        Args:
            signal_names (list(str)): List of signal names to query

        Returns:
            list(tuple((str), (str), (int), (int), (int), (int))):
                name (str): The signal name specified by the caller

                description (str): A human-readable description of
                                   what the signal measures.

                domain_type (int): One of the geopmpy.topo.DOMAIN_*
                                   integers corresponding to a domain
                                   type where the signal is natively
                                   supported.

                aggregation (int): One of the geopm::Agg::m_type_e
                                   enum values defined in Agg.hpp that
                                   specifies how to aggregate the
                                   signal over domains.

                format_type (int): One of the geopm::string_format_e
                                   enum values defined in Helper.hpp
                                   that specifies how to convert the
                                   signal value to a human readable
                                   string.

                behavior (int): One of the
                                IOGroup::m_signal_behavior_e enum
                                values that specifies how the signal
                                changes over time.

        """
        result = []
        for name in signal_names:
            description = self._pio.signal_description(name)
            domain_type = self._pio.signal_domain_type(name)
            aggregation, format_type, behavior = self._pio.signal_info(name)
            result.append((name, description, domain_type, aggregation, format_type, behavior))
        return result

    def get_control_info(self, control_names):
        """For each specified control name, return a tuple of information.

        The caller provides a list of control names.  This method returns
        a list of the same length which consists of a tuple of information
        for each control.

        The method will raise a RuntimeError if any of the requested
        control names are not supported by the service.

        Args:
            control_names (list(str)): List of control names to query

        Returns:
            list(tuple((str), (str), (int))):
                name (str): The control name specified by the caller

                description (str): A human-readable description of the
                                   effect of setting the control.

                domain_type (int): One of the geopmpy.topo.DOMAIN_*
                                   integers corresponding to a domain
                                   type that is the finest available
                                   granularity for the control.

        """
        result = []
        for name in control_names:
            description = self._pio.control_description(name)
            domain_type = self._pio.control_domain_type(name)
            result.append((name, description, domain_type))
        return result

    def lock_control(self):
        """Block all write-mode sessions.

        A call to this method will end any currently running
        write-mode session and block the creation of any new
        write-mode sessions.  Use the unlock_control() method to
        enable write-mode sessions to begin again.

        Calls to this method through the DBus interface will be
        blocked for all non-root users due to the rules established by
        the "io.github.geopm.conf" DBus configuration file.

        Until write-mode sessions are re-enabled, any calls to
        write_control() or calls to start_batch() with a non-empty
        control_config parameter will raise a RuntimeError.

        The default system configuration may be updated by the system
        administrator while write-mode is locked.  For instance, the
        system administrator may use the geopmwrite command-line tool
        as user root between calls to lock_control() and
        unlock_control().  These changes will be reflected by the
        save/restore executed by the service for the next client
        write-mode session after the controls are unlocked.

        No action is taken and no error is raised if controls are already
        disabled when this method is called.

        """
        raise NotImplementedError('PlatformService: Implementation incomplete')

    def unlock_control(self):
        """Unblock access to create new write-mode sessions.

        A call to this method will re-enable write-mode sessions to be
        created after they were previously disabled by a call to
        lock_control().

        Although new sessions may be created after a call to
        unlock_control(), write-mode sessions that were ended due to
        previous calls to lock_control() will remain closed.

        No action is taken and no error is raised if controls are already
        enabled when this method is called.

        """
        raise NotImplementedError('PlatformService: Implementation incomplete')

    def open_session(self, user, client_pid):
        """Open a new session for the client thread.

        The creation of a session is the first step that a client
        thread makes to interact with the GEOPM service.  Each Linux
        thread has a unique PID, and this PID can be associated with
        at most one GEOPM service session.  The session is ended by
        passing the client PID to the close_session() method.  The
        session will also be ended if the client thread terminates for
        any reason.

        The client PID may attempt to open a new session multiple
        times.  In this case a reference count records the number of
        calls to open the session by the client PID.  Each call to
        open a session is expected to be paired with a call to close
        the session.  When the reference count falls to zero, all
        resources associated with the session are released.

        Incrementing the reference count is an abstraction for the
        client.  The client itself cannot have more than a single open
        connection at once, so when requesting a new session, other
        than incrementing the counter, nothing else happens.

        All sessions are opened in read-mode, and may later be
        promoted to write-mode.  This promotion will occur with the
        first successful call requiring access to controls.  This
        includes any calls to write_control() or calls to
        start_batch() with a non-empty control_config.  These calls
        will fail if there is an active write-mode session opened by
        another thread.

        Prior to being promoted to write-mode, the current value for
        all controls that are supported by the GEOPM service will be
        recorded.  This record enables the service to restore all
        controls when the write-mode session ends.

        Permissions for the session's access to signals and controls
        are determined based on the policy for the user when
        the session is first opened.  Calling the set_group_access()
        method to alter the policy will not affect active sessions.

        Args:
            user (str): Unix user name that owns the client thread
                        opening the session.

            client_pid (int): Linux PID of the client thread opening
                              the session.

        """
        if self._active_sessions.is_client_active(client_pid):
            self._active_sessions.increment_reference_count(client_pid)
        else:
            signals, controls = self.get_user_access(user)
            watch_id = self._watch_client(client_pid)
            self._active_sessions.add_client(client_pid, signals, controls, watch_id)

    def close_session(self, client_pid):
        """Close an active session for the client process.

        After closing a session, the client process is required to
        call open_session() again before using any of the client-facing
        member functions.

        Closing an active session will remove the record of which
        signals and controls the client process has access to; thus,
        this record is updated to reflect changes to the policy when the
        next session is opened by the process.

        When closing a write-mode session, the control values that were
        recorded when the session was promoted to write-mode are restored.

        A client PID may attempt to open a session multiple times and
        this is tracked with a reference count as described in the
        documentation for opening a session.  In this case, calls to
        close the session will not actually release the resources
        associated with the session until the reference count falls to
        zero.

        A RuntimeError is raised if the client_pid does not have an
        open session.

        Args:
            client_pid (int): Linux PID of the client thread

        """
        self._active_sessions.check_client_active(client_pid, 'PlatformCloseSession')
        reference_count = self._active_sessions.get_reference_count(client_pid)
        if reference_count == 0:
            # The daemon died, restarted, and found a session with
            # zero reference count.
            self._close_session_completely(client_pid)
        elif reference_count == 1:
            self._active_sessions.decrement_reference_count(client_pid)
            self._close_session_completely(client_pid)
        else:  # reference_count > 1:
            self._active_sessions.decrement_reference_count(client_pid)

    def close_session_admin(self, client_pid):
        """Close an active session for the client process completely.

        This administrative function is used to forcibly close an
        active session regardless of the client's reference count.
        Similarly, if a client process PID is no longer active, but an
        open session exists, this implementation is used to release
        the session resources.

        When closing a write-mode session, the control values that were
        recorded when the session was promoted to write-mode are restored.
        A RuntimeError is raised if the client_pid does not have an
        open session.

        Args:
            client_pid (int): Linux PID of the client thread

        """
        self._close_session_completely(client_pid)

    def _close_session_completely(self, client_pid):
        """Close an active session for the client process completely.

        This method is called internally by close_session() and it is
        also called explicitly by check_client() when the process dies.

        A RuntimeError is raised if the client_pid does not have an
        open session.

        Args:
            client_pid (int): Linux PID of the client thread

        """
        sess = self._active_sessions.remove_client(client_pid)
        if sess is not None:
            GLib.source_remove(sess['watch_id'])
            batch_pid = sess.get('batch_server')
            if batch_pid is not None:
                try:
                    self._pio.stop_batch_server(batch_pid)
                except RuntimeError:
                    sys.stderr.write(f"Warning: Failed to call pio.stop_batch_server({batch_pid}), sending SIGKILL\n)")
                    if psutil.pid_exists(batch_pid):
                        os.kill(batch_pid, signal.SIGKILL)
            with system_files.WriteLock(self._RUN_PATH) as lock:
                if lock.try_lock() == client_pid:
                    self._close_session_write(lock, client_pid)

    def _close_session_write(self, lock, pid):
        save_dir = os.path.join(self._RUN_PATH, self._SAVE_DIR)
        if os.path.isdir(save_dir):
            is_restored = False
            try:
                self._pio.restore_control_dir(save_dir)
                is_restored = True
            except RuntimeError as ex:
                sys.stderr.write(f'Failed to restore control settings for client {pid}: {ex}')
            del_dir = f'{save_dir}-{pid}-{uuid.uuid4()}-del'
            os.rename(save_dir, del_dir)
            if is_restored:
                shutil.rmtree(del_dir)
            else:
                sys.stderr.write(f'Failed to restore controls for PID {pid}, moved to {del_dir}')
        else:
            sys.stderr.write(f'Failed to restore controls for PID {pid}, {save_dir} is not a directory')
        lock.unlock(pid)

    def start_batch(self, client_pid, signal_config, control_config):
        """Start a batch server to support a client session.

        Configure the signals and controls that will be enabled by the
        batch server and start the server process.  The server enables
        fast access for the signals and controls that are configured
        by the caller.  These are configured by specifying a name,
        domain and domain index for each of the signals and controls
        that the server will support.

        After a batch server is successfully created, the client will
        interact with the batch server though PlatformIO interfaces
        that do not go over DBus.  That is, once access is established
        by DBus, a faster protocol can be safely used.

        The batch server does not enable features beyond those of
        the read_signal() or write_control() methods; it simply
        provides a much higher performance interface.

        A RuntimeError is raised if the client does not have
        permission to read or write any of configured signals or
        controls or if the client_pid does not have an open session.
        The server process will not be created if any error occurs.

        Args:
            client_pid (int): Linux PID of the client thread

            signal_config (list(tuple((int), (int), (str))):
                domain_type (int): One of the geopmpy.topo.DOMAIN_*
                                   integers corresponding to a domain
                                   type to read from.

                domain_idx (int): Specifies the particular domain
                                  index to read from.

                signal_name (str): The name of the signal to read.

            control_config (list(tuple((int), (int), (str))):
                domain_type (int): One of the geopmpy.topo.DOMAIN_*
                                   integers corresponding to a domain
                                   type to write to.

                domain_idx (int): Specifies the particular domain
                                  index to write to.

                control_name (str): The name of the control to write.

        Returns:
            tuple(int, str):
                server_pid (int): The Linux PID of the batch server process.

                server_key (str): A unique identifier enabling the
                                  server/client connection across
                                  inter-process shared memory.

        """
        self._active_sessions.check_client_active(client_pid, 'PlatformStartBatch')
        sig_req = {cc[2] for cc in signal_config}
        cont_req = {cc[2] for cc in control_config}
        supported_signals = self._active_sessions.get_signals(client_pid)
        supported_controls = self._active_sessions.get_controls(client_pid)
        if not sig_req.issubset(supported_signals):
            raise RuntimeError('Requested signals that are not in allowed list: {}' \
                               .format(sorted(sig_req.difference(supported_signals))))
        elif not cont_req.issubset(supported_controls):
            raise RuntimeError('Requested controls that are not in allowed list: {}' \
                               .format(sorted(cont_req.difference(supported_controls))))
        if len(control_config) != 0:
            self._write_mode(client_pid)
        batch_pid = self._active_sessions.get_batch_server(client_pid)
        if batch_pid is not None:
            raise RuntimeError(f'Client {client_pid} has already started a batch server: {batch_pid}')
        batch_pid , batch_key = self._pio.start_batch_server(client_pid, signal_config, control_config)
        self._active_sessions.set_batch_server(client_pid, batch_pid)
        return batch_pid, batch_key

    def stop_batch(self, client_pid, server_pid):
        """End a batch server previously started by the client.

        Terminate the batch server process and free up all resources
        associated with the batch server.  Any future calls by the
        client to interfaces that require the batch server will result
        in errors.

        The batch server will also be terminated if the client session
        that created the server terminates for any reason.

        A RuntimeError will be raised if the specified server PID was
        not previously created by the client's call to start_batch(),
        or if the batch server has already been closed for any reason.

        Args:
            client_pid (int): Linux PID of the client thread.

            server_pid (int): Linux PID of a batch server process
                              returned by a previous call to
                              start_batch().

        """
        self._active_sessions.check_client_active(client_pid, 'StopBatch')
        actual_server_pid = self._active_sessions.get_batch_server(client_pid)
        if server_pid != actual_server_pid:
            raise RuntimeError(f'Client PID: {client_pid} requested to stop batch server PID: {server_pid}, actual batch server PID: {actual_server_pid}')
        self._pio.stop_batch_server(server_pid)
        self._active_sessions.remove_batch_server(client_pid)

    def read_signal(self, client_pid, signal_name, domain, domain_idx):
        """Read a signal from a particular domain.

        Select a signal by name and read the current value from the
        domain type and index specified.  The value is returned
        as a floating point number in SI units.

        A RuntimeError is raised if the client_pid does not have an
        open session or if the client does not have permission to read the
        signal from the specified domain.

        A RuntimeError is raised if the requested signal is not
        supported, the domain is invalid or the domain index is out of
        range.

        Args:
            client_pid (int): Linux PID of the client thread.

            signal_name (str): Name of the signal to read.

            domain (int): One of the geopmpy.topo.DOMAIN_* integers
                          corresponding to a domain type to read from.

            domain_idx (int): Specifies the particular domain index to
                              read from.

        Returns:
            (float): The value of the signal in SI units.

        """
        self._active_sessions.check_client_active(client_pid, 'PlatformReadSignal')
        signal_avail = self._active_sessions.get_signals(client_pid)
        if not signal_name in signal_avail:
            raise RuntimeError('Requested signal that is not in allowed list: {}'.format(signal_name))
        return self._pio.read_signal(signal_name, domain, domain_idx)

    def write_control(self, client_pid, control_name, domain, domain_idx, setting):
        """Write a control value to a particular domain.

        Select a control by name and write a new setting for the
        domain type and index specified.  The written control setting
        is a floating point number in SI units.

        A RuntimeError is raised if the client_pid does not have an
        open session, or if the client does not have permission to write the
        control to the specified domain.

        A RuntimeError is raised if a different client currently has an
        open write-mode session.

        A RuntimeError is raised if the requested control is not
        supported, the domain is invalid or the domain index is out of
        range.

        Args:
            client_pid (int): Linux PID of the client thread.

            control_name (str): The name of the control to write.

            domain (int): One of the geopmpy.topo.DOMAIN_* integers
                          corresponding to a domain type to written
                          to.

            domain_idx (int): Specifies the particular domain index to
                              write to.

            setting (float): Value of the control to be written.

        """


        self._active_sessions.check_client_active(client_pid, 'PlatformWriteControl')
        control_avail = self._active_sessions.get_controls(client_pid)
        if not control_name in control_avail:
            raise RuntimeError('Requested control that is not in allowed list: {}'.format(control_name))
        self._write_mode(client_pid)
        self._pio.write_control(control_name, domain, domain_idx, setting)

    def _write_mode(self, client_pid):
        write_pid = client_pid
        do_open_session = False
        # If the session leader is an active process then tie the write lock
        # to the session leader, otherwise the write lock is associated with
        # the requesting process.
        client_sid = os.getsid(client_pid)
        if client_sid != client_pid and psutil.pid_exists(client_sid):
            write_pid = client_sid
            do_open_session = True
        with system_files.WriteLock(self._RUN_PATH) as lock:
            lock_pid = lock.try_lock()
            if lock_pid is None:
                if do_open_session:
                    # If the write lock is associated with the session leader
                    # process, then open a session for the session leader
                    session_uid = os.stat(f'/proc/{write_pid}/status').st_uid
                    session_user = pwd.getpwuid(session_uid).pw_name
                    self.open_session(session_user, write_pid)
                save_dir = os.path.join(self._RUN_PATH, self._SAVE_DIR)
                # Clean up existing SAVE_FILES directory if it exists
                if os.path.exists(save_dir):
                    del_dir = f'{save_dir}-{uuid.uuid4()}-del'
                    sys.stderr.write(f'Warning: <geopm-service> Renaming SAVE_FILES directory which not owned by any active session to {del_dir}\n')
                    os.rename(save_dir, del_dir)
                # Save Control values
                tmp_dir = f'{save_dir}-{uuid.uuid4()}-tmp'
                system_files.secure_make_dirs(tmp_dir)
                self._pio.save_control_dir(tmp_dir)
                os.rename(tmp_dir, save_dir)
                # Set the write lock to the writer PID
                lock.try_lock(write_pid)
            elif lock_pid != write_pid:
                raise RuntimeError(f'The PID {client_pid} requested write access, but the geopm service already has write mode client with PID or SID of {lock_pid}')

    def _watch_client(self, client_pid):
        return GLib.timeout_add_seconds(self._WATCH_INTERVAL_SEC, self.check_client, client_pid)

    def check_client(self, client_pid):
        """Called by GLib periodically to monitor if a PID is active

        GLib queries check_client() to see if the process is still alive.

        This method gets triggered upon abnormal termination of the
        session, such as when the client process unexpectedly crashes
        or ends without closing all sessions.

        """
        if (client_pid in self._active_sessions.get_clients() and
            not psutil.pid_exists(client_pid)):
            self._close_session_completely(client_pid)
            return False
        return True


class TopoService(object):
    """Provides the concrete implementation for all of the GEOPM DBus
    interfaces that use PlatformTopo.  This class is used by the
    GEOPMService class that maps DBus interfaces to their
    implementations.

    """
    def __init__(self, topo=topo):
        """TopoService constructor that initializes all private members.

        """
        self._topo = topo

    def get_cache(self):
        """Return the contents of the PlatformTopo cache file.

        Create the PlatformTopo cache file if it does not exist and
        then return the contents of the file as a string.  This
        provides all the information required to associate all domains
        with CPUs.

        Returns:
            (str): Contents of the topology cache file that defines
                   the system topology.

        """
        self._topo.create_cache()
        with open('/run/geopm-service/geopm-topo-cache') as fid:
            result = fid.read()
        return result


class GEOPMService(object):
    """The dasbus service object that is published.

    Object used by dasbus to map GEOPM service DBus APIs to their
    implementation.  A GEOPMService object is published by geopmd
    using the publish_object() method of dasbus.SystemMessageBus
    class which enables the GEOPM service.

    The GEOPMService methods are named after the DBus API they provide
    within the io.github.geopm DBus namespace.  The parameter type
    conversion is configured with the __dbus_xml__ class member.

    The implementation for each method is a pass-through call to a
    member object method.  These member objects do not have an
    explicit dasbus dependency which facilitates unit testing.

    """
    __dbus_xml__ = dbus_xml.geopm_dbus_xml(TopoService, PlatformService)

    def __init__(self):
        self._topo = TopoService()
        self._platform = PlatformService()
        self._dbus_proxy = SystemMessageBus().get_proxy('org.freedesktop.DBus',
                                                        '/org/freedesktop/DBus')

    def TopoGetCache(self):
        return self._topo.get_cache()

    def PlatformGetGroupAccess(self, group):
        return self._platform.get_group_access(group)

    @accepts_additional_arguments
    def PlatformSetGroupAccess(self, group, allowed_signals, allowed_controls, **call_info):
        self._check_cap_sys_admin(call_info, "PlatformSetGroupAccess")
        self._platform.set_group_access(group, allowed_signals, allowed_controls)

    @accepts_additional_arguments
    def PlatformGetUserAccess(self, **call_info):
        return self._platform.get_user_access(self._get_user(**call_info))

    def PlatformGetAllAccess(self):
        return self._platform.get_all_access()

    def PlatformGetSignalInfo(self, signal_names):
        return self._platform.get_signal_info(signal_names)

    def PlatformGetControlInfo(self, control_names):
        return self._platform.get_control_info(control_names)

    @accepts_additional_arguments
    def PlatformLockControl(self, **call_info):
        self._check_cap_sys_admin(call_info, "PlatformLockControl")
        self._platform.lock_control()

    @accepts_additional_arguments
    def PlatformUnlockControl(self, **call_info):
        self._check_cap_sys_admin(call_info, "PlatformUnLockControl")
        self._platform.unlock_control()

    @accepts_additional_arguments
    def PlatformOpenSession(self, **call_info):
        self._platform.open_session(self._get_user(**call_info), self._get_pid(**call_info))

    @accepts_additional_arguments
    def PlatformCloseSession(self, **call_info):
        self._platform.close_session(self._get_pid(**call_info))

    @accepts_additional_arguments
    def PlatformCloseSessionAdmin(self, client_pid, **call_info):
        self._check_cap_sys_admin(call_info, "PlatformCloseSessionAdmin")
        self._platform.close_session_admin(client_pid)

    @accepts_additional_arguments
    def PlatformStartBatch(self, signal_config, control_config, **call_info):
        return self._platform.start_batch(self._get_pid(**call_info), signal_config, control_config)

    @accepts_additional_arguments
    def PlatformStopBatch(self, server_pid, **call_info):
        self._platform.stop_batch(self._get_pid(**call_info), server_pid)

    @accepts_additional_arguments
    def PlatformReadSignal(self, signal_name, domain, domain_idx, **call_info):
        return self._platform.read_signal(self._get_pid(**call_info), signal_name, domain, domain_idx)

    @accepts_additional_arguments
    def PlatformWriteControl(self, control_name, domain, domain_idx, setting, **call_info):
        self._platform.write_control(self._get_pid(**call_info), control_name, domain, domain_idx, setting)

    def _get_user(self, call_info):
        """Use DBus proxy object to derive the user name that owns the client
           process.

        Args:
           call_info (dict): The dictionary provided by the dasbus
                             @accepts_additional_arguments decorator.

        Returns:
            (str): The Unix user name of the client process owner.

        """
        unique_name = call_info['sender']
        uid = self._dbus_proxy.GetConnectionUnixUser(unique_name)
        return pwd.getpwuid(uid).pw_name

    def _get_pid(self, call_info):
        """Use DBus proxy object to derive the Linux PID of the client
           thread.

        Args:
           call_info (dict): The dictionary provided by the dasbus
                             @accepts_additional_arguments decorator.

        Returns:
            (int): The Linux PID of the client thread.

        """
        unique_name = call_info['sender']
        return self._dbus_proxy.GetConnectionUnixProcessID(unique_name)

    def _check_cap_sys_admin(self, call_info, api_name):
        has_admin = False
        cap = 0
        cap_sys_admin = 0x00200000
        pid = self._get_pid(**call_info)
        status_path = f'/proc/{pid}/status'
        with open(status_path) as fid:
            for line in fid.readlines():
                if line.startswith('CapEff:'):
                    cap = int(line.split(':')[1], 16)
        if cap & cap_sys_admin == 0:
            raise RuntimeError('Calling "io.github.geopm.{api_name}" failed, try with sudo or as "root" user (requires CAP_SYS_ADMIN)')



#!/usr/bin/env python3

#  Copyright (c) 2015 - 2022, Intel Corporation
#  SPDX-License-Identifier: BSD-3-Clause
#

from setuptools import setup

import os
import sys
import shutil

from setuptools import setup

if os.getcwd() != os.path.dirname(os.path.abspath(__file__)):
    sys.stderr.write('ERROR:  script must be run in the directory that contains it\n')
    sys.exit(1)

try:
    from geopmdpy import __version__
except IOError:
    sys.stderr.write('WARNING:  geopmpy/version.py not found, setting version to 0.0.0\n')
    __version__ = '0.0.0'

long_description = """\
The python implementation for the GEOPM daemon"""


scripts = ['geopmd', 'geopmaccess', 'geopmsession']

classifiers = ['Development Status :: 5 - Production/Stable',
               'License :: OSI Approved :: BSD License',
               'Operating System :: POSIX :: Linux',
               'Natural Language :: English',
               'Topic :: Scientific/Engineering',
               'Topic :: Software Development :: Libraries :: Application Frameworks',
               'Topic :: System :: Hardware :: Symmetric Multi-processing',
               'Topic :: System :: Power (UPS)',
               'Programming Language :: Python :: 3',
               'Programming Language :: Python :: 3.6',
]

install_requires = ['cffi>=1.15.0',
                    'setuptools>=59.6.0',
                    'psutil>=5.9.0',
                    'dasbus>=1.6.0',
                    'jsonschema>=3.2.0']

setup(name='geopmdpy',
      version=__version__,
      description='GEOPM - Global Extensible Open Power Manager Daemon',
      long_description=long_description,
      url='https://geopm.github.io',
      download_url='https://github.com/geopm/geopm/releases',
      license='BSD-3-Clause',
      author='Christopher Cantalupo <christopher.m.cantalupo@intel.com>',
      packages=['geopmdpy'],
      scripts=scripts,
      classifiers=classifiers,
      install_requires=install_requires,
      python_requires='>=3.6')



#  Copyright (c) 2015 - 2022, Intel Corporation
#  SPDX-License-Identifier: BSD-3-Clause
#

"""Tools to support runtime implementations that use the GEOPM service

"""

import math
import time
import subprocess # nosec
import sys
import shlex
from . import pio


class TimedLoop:
    """Object that can be iterated over to run a timed loop

    Use in a for loop to execute a fixed number of timed delays.  The
    overhead time for executing what is inside of the loop is
    accounted for.  Calls to time.sleep() are made to delay until the
    targeted end time for each iteration.

    Example:

        >>> from time import time
        >>> time_0 = time()
        >>> for index in TimedLoop(0.1, 10):
        ...     print(f'{index}: {time() - time_0}')
        ...
        0: 0.0008680820465087891
        1: 0.10126090049743652
        2: 0.20174455642700195
        3: 0.30123186111450195
        4: 0.4010961055755615
        5: 0.5020360946655273
        6: 0.6011238098144531
        7: 0.7011349201202393
        8: 0.8020164966583252
        9: 0.9015650749206543
        10: 1.0021190643310547

    """

    def __init__(self, period, num_period=None):
        """Constructor for timed loop object

        The number of loops executed is one greater than the number of
        time intervals requested, and that the first iteration is not
        delayed.  The total amount of time spanned by the loop is the
        product of the two input parameters.

        To create an infinite loop, specify num_period is None.

        Args:
            period (float): Target interval for the loop execution in
                            units of seconds.

            num_period (int): Number of time periods spanned by the
                              loop.  The total loop time is
                              num_periods * period, but since there is
                              no delay in the first loop, there will
                              be num_period + 1 loop iterations.

        """

        if period < 0.0:
            raise RuntimeError('Specified period is invalid.  Must be >= 0.')
        if num_period is not None:
            if num_period < 0:
                raise RuntimeError('Specified num_period is invalid.  Must be > 0.')
            if not isinstance(num_period, int):
                raise ValueError('num_period must be a whole number.')
        self._period = period
        self._num_loop = num_period
        # Add one to ensure:
        #     total_time == num_loop * period
        # because we do not delay the start iteration
        if self._num_loop is not None:
            self._num_loop += 1

    def __iter__(self):
        """Set up a timed loop

        Iteration method for timed loop.  This iterator can be used in
        a for statement to execute the loop periodically.

        """
        self._loop_idx = 0
        self._target_time = time.time()
        return self

    def __next__(self):
        """Sleep until next targeted time for loop and update counter

        """
        result = self._loop_idx
        if self._loop_idx == self._num_loop:
            raise StopIteration
        if self._loop_idx != 0:
            sleep_time = self._target_time - time.time()
            if sleep_time > 0:
                self.wait(sleep_time)
        self._target_time += self._period
        self._loop_idx += 1
        return result

    def wait(self, timeout):
        """Pass-through to time.sleep()

        Args:
            timeout (float): Target interval for the loop execution in
                             units of seconds.
        """
        time.sleep(timeout)


class PIDTimedLoop(TimedLoop):
    def __init__(self, pid, period, num_period=None):
        """Similar to the TimedLoop but stop when subprocess ends

        The number of loops executed is one greater than the number of
        time intervals requested, and that the first iteration is not
        delayed.  The total amount of time spanned by the loop is the
        product of the two input parameters.

        To create an infinite loop, specify num_period is None.

        Loop will always terminate when the subprocess pid terminates.

        Args:
            pid (Popen): Object returned by subprocess.Popen() constructor.

            period (float): Target interval for the loop execution in
                            units of seconds.

            num_period (int): Number of time periods spanned by the
                              loop.  The total loop time is
                              num_periods * period, but since there is
                              no delay in the first loop, there will
                              be num_period + 1 loop iterations.

        """

        super(PIDTimedLoop, self).__init__(period, num_period)
        self._pid = pid
        self._is_active = pid.poll() is None

    def wait(self, timeout):
        """Wait for timeout seconds or until pid ends

        Args:
            timeout (float): Target interval for the loop execution in
                             units of seconds.

        Raises:
            StopIteration: When last call to wait termintated due to
                           the process ending

        """
        if not self._is_active:
            raise StopIteration
        try:
            self._pid.wait(timeout=timeout)
            self._is_active = False
        except subprocess.TimeoutExpired:
            pass

class Agent:
    """Base class that documents the interfaces required by an agent

    Agent objects are used to initialize a Controller object and
    define the control algorithm.

    """
    def __init__(self):
        raise NotImplementedError('Agent is an abstract base class')

    def get_signals(self):
        """Get list of read requests

        The returned signals will be sampled from the platform by the
        Controller and passed into Agent.update().

        Returns:
            list((str, int, int)): List of request tuples. Each
                                   request comprises a signal name,
                                   domain type, and domain index.

        """
        raise NotImplementedError('Agent is an abstract base class')

    def get_controls(self):
        """Get list of control requests

        The returned controls will be set in the platform by the Controller
        based on the return value from Agent.update().

        Returns:
            list((str, int, int)): List of request tuples. Each
                                   request comprises a control name,
                                   domain type, and domain index.

        """
        raise NotImplementedError('Agent is an abstract base class')

    def run_begin(self, policy, profile):
        """Called by Controller at the start of each run

        The policy for the run is passed through to the agent from the
        Controller.run() input.  For some agents, the policy may always be
        None.

        Args:
            policy (object): The Agent specific policy provided to the
                             Controller.run() method.

            profile (str): Profile name to associate with the report

        """
        raise NotImplementedError('Agent is an abstract base class')

    def policy_repr(self, policy):
        """Create a string representation of a policy suitable for printing

        """
        return policy.__repr__()

    def run_end(self):
        """Called by the Controller at the end of each run

        The result of calling the get_report() method after run_end() should
        reflect the same report until the next call to run_end().  This report
        will document the measurements made between the last calls to
        run_begin() and run_end().  Each call to run_end() will follow a
        previous call to run_begin().  The run_end() method will be called by
        the Controller even if the run resulted in an error that raises an
        exception.  In this way resources associated with a single run can be
        released when the run_end() method is called.

        """
        raise NotImplementedError('Agent is an abstract base class')

    def update(self, signals):
        """Called periodically by the Controller

        The signals that specified by get_signals() will be passed as inputs
        to the method by the Controller.  The update() method will be called
        periodically and the interval is set by the value returned by
        Agent.get_period().

        Args:
            signals (list(float)): Recently read signal values

        Returns:
            list(float): Control values for next control interval

        """
        raise NotImplementedError('Agent is an abstract base class')

    def get_period(self):
        """Get the target time interval for the control loop

        Returns:
            float: Time interval in seconds

        """
        raise NotImplementedError('Agent is an abstract base class')

    def get_report(self):
        """Summary of all data collected by calls to update()

        The report covers the interval of time between the last two calls to
        Agent.begin_run() / Agent.end_run().  Until the next call to
        Agent.begin_run(), the same report will be returned by this method.
        The Controller.run() method will return this report upon completion of
        the run.

        Returns:
            str: Human readable report

        """
        raise NotImplementedError('Agent is an abstract base class')


class Controller:
    """Class that supports a runtime control algorithm

    """
    def __init__(self, agent, timeout=0):
        """Controller constructor

        Args:
            agent (Agent): Object that conforms to the Agent class
                           interface

            timeout (float): The agent algorithm will run for the full
                             duration of the application execution if timeout
                             is 0.  Setting the timeout to a non-zero value
                             will end the agent algorithm after the specified
                             period of time or when the application ends,
                             whichever occurs first.

        """
        if not isinstance(agent, Agent):
            raise ValueError('agent must be a subclass of Agent.')
        if timeout < 0:
            raise ValueError('timeout must be >= 0')
        self._agent = agent
        self._signals = agent.get_signals()
        self._controls = agent.get_controls()
        self._signals_idx = []
        self._controls_idx = []
        self._update_period = agent.get_period()
        self._num_update = None
        if timeout != 0:
            self._num_update = math.ceil(timeout / self._update_period)
        self._returncode = None

    def push_all(self):
        self._signals_idx = [pio.push_signal(*ss) for ss in self._signals]
        self._controls_idx = [pio.push_control(*cc) for cc in self._controls]

    def read_all_signals(self):
        """Sample for all signals pushed with pio

        Returns:
            list(float): Sampled values for each signal

        """
        return [pio.sample(signal_idx)
                for signal_idx in self._signals_idx]

    def returncode(self):
        """Get the return code of the application process

        Returns:
            int: Return code of app process
        """
        if self._returncode is None:
            raise RuntimeError('App process is still running')
        return self._returncode

    def run(self, argv, policy=None, profile=None):
        """Execute control loop defined by agent

        Interfaces with PlatformIO directly through the geopmdpy.pio module.

        Args:
            argv (list(str)): Arguments for application that is executed

            policy (object): Policy for Agent to use during the run

        Returns:
            str: Human readable report created by agent

        """
        sys.stderr.write('<geopmdpy> RUN BEGIN\n')
        if profile is None:
            profile = ' '.join([shlex.quote(arg) for arg in argv])
        try:
            pio.save_control()
            self.push_all()
            pid = subprocess.Popen(argv)
            self._agent.run_begin(policy, profile)
            for loop_idx in PIDTimedLoop(pid, self._update_period, self._num_update):
                pio.read_batch()
                signals = self.read_all_signals()
                new_settings = self._agent.update(signals)
                if pid.poll() is not None:
                    break
                for control_idx, new_setting in zip(self._controls_idx, new_settings):
                    pio.adjust(control_idx, new_setting)
                pio.write_batch()
            self._agent.run_end()
        except:
            raise
        finally:
            pio.restore_control()
        self._returncode = pid.returncode
        sys.stderr.write(f'<geopmdpy> RUN END, return: {self._returncode}\n')
        return self._agent.get_report()



#  Copyright (c) 2015 - 2022, Intel Corporation
#  SPDX-License-Identifier: BSD-3-Clause
#

from dasbus.loop import EventLoop
from dasbus.connection import SystemMessageBus
from signal import signal
from signal import SIGTERM
from . import service

_bus = None

def term_handler(signum, frame):
    if signum == SIGTERM:
        stop()

def stop():
    if _bus is not None:
        _bus.disconnect()
    exit(0)

def main():
    signal(SIGTERM, term_handler)
    loop = EventLoop()
    global _bus
    _bus = SystemMessageBus()
    try:
        _bus.publish_object("/io/github/geopm", service.GEOPMService())
        _bus.register_service("io.github.geopm")
        loop.run()
    finally:
        stop()

if __name__ == '__main__':
    main()
